from django.test import TestCase

from .models import MicroBlogPost


class AppConfigTests(TestCase):
    def test_index_collection(self):
        from haystack import connections

        unified_index = connections["default"].get_unified_index()
        models = unified_index.get_indexed_models()

        self.assertIn(MicroBlogPost, models)



# A couple models for Haystack to test with.
import datetime
import uuid

from django.db import models


class MockTag(models.Model):
    name = models.CharField(max_length=32)

    def __str__(self):
        return self.name


class MockModel(models.Model):
    author = models.CharField(max_length=255)
    foo = models.CharField(max_length=255, blank=True)
    pub_date = models.DateTimeField(default=datetime.datetime.now)
    tag = models.ForeignKey(MockTag, models.CASCADE)

    def __str__(self):
        return self.author

    def hello(self):
        return "World!"


class UUIDMockModel(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    characteristics = models.TextField()

    def __str__(self):
        return str(self.id)


class AnotherMockModel(models.Model):
    author = models.CharField(max_length=255)
    pub_date = models.DateTimeField(default=datetime.datetime.now)

    def __str__(self):
        return self.author


class AThirdMockModel(AnotherMockModel):
    average_delay = models.FloatField(default=0.0)
    view_count = models.PositiveIntegerField(default=0)


class CharPKMockModel(models.Model):
    key = models.CharField(primary_key=True, max_length=10)


class AFourthMockModel(models.Model):
    author = models.CharField(max_length=255)
    editor = models.CharField(max_length=255)
    pub_date = models.DateTimeField(default=datetime.datetime.now)

    def __str__(self):
        return self.author


class SoftDeleteManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(deleted=False)

    def complete_set(self):
        return super().get_queryset()


class AFifthMockModel(models.Model):
    author = models.CharField(max_length=255)
    deleted = models.BooleanField(default=False)

    objects = SoftDeleteManager()

    def __str__(self):
        return self.author


class ASixthMockModel(models.Model):
    name = models.CharField(max_length=255)
    lat = models.FloatField()
    lon = models.FloatField()

    def __str__(self):
        return self.name


class ScoreMockModel(models.Model):
    score = models.CharField(max_length=10)

    def __str__(self):
        return self.score


class ManyToManyLeftSideModel(models.Model):
    related_models = models.ManyToManyField("ManyToManyRightSideModel")


class ManyToManyRightSideModel(models.Model):
    name = models.CharField(max_length=32, default="Default name")

    def __str__(self):
        return self.name


class OneToManyLeftSideModel(models.Model):
    pass


class OneToManyRightSideModel(models.Model):
    left_side = models.ForeignKey(
        OneToManyLeftSideModel, models.CASCADE, related_name="right_side"
    )



import datetime

from django.contrib.gis.measure import D
from django.test import TestCase

from haystack import connections
from haystack.inputs import Exact
from haystack.models import SearchResult
from haystack.query import SQ, SearchQuerySet

from ..core.models import AnotherMockModel, MockModel


class Elasticsearch5SearchQueryTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.sq = connections["elasticsearch"].get_query()

    def test_build_query_all(self):
        self.assertEqual(self.sq.build_query(), "*:*")

    def test_build_query_single_word(self):
        self.sq.add_filter(SQ(content="hello"))
        self.assertEqual(self.sq.build_query(), "(hello)")

    def test_build_query_boolean(self):
        self.sq.add_filter(SQ(content=True))
        self.assertEqual(self.sq.build_query(), "(True)")

    def test_regression_slash_search(self):
        self.sq.add_filter(SQ(content="hello/"))
        self.assertEqual(self.sq.build_query(), "(hello\\/)")

    def test_build_query_datetime(self):
        self.sq.add_filter(SQ(content=datetime.datetime(2009, 5, 8, 11, 28)))
        self.assertEqual(self.sq.build_query(), "(2009-05-08T11:28:00)")

    def test_build_query_multiple_words_and(self):
        self.sq.add_filter(SQ(content="hello"))
        self.sq.add_filter(SQ(content="world"))
        self.assertEqual(self.sq.build_query(), "((hello) AND (world))")

    def test_build_query_multiple_words_not(self):
        self.sq.add_filter(~SQ(content="hello"))
        self.sq.add_filter(~SQ(content="world"))
        self.assertEqual(self.sq.build_query(), "(NOT ((hello)) AND NOT ((world)))")

    def test_build_query_multiple_words_or(self):
        self.sq.add_filter(~SQ(content="hello"))
        self.sq.add_filter(SQ(content="hello"), use_or=True)
        self.assertEqual(self.sq.build_query(), "(NOT ((hello)) OR (hello))")

    def test_build_query_multiple_words_mixed(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(content="hello"), use_or=True)
        self.sq.add_filter(~SQ(content="world"))
        self.assertEqual(
            self.sq.build_query(), "(((why) OR (hello)) AND NOT ((world)))"
        )

    def test_build_query_phrase(self):
        self.sq.add_filter(SQ(content="hello world"))
        self.assertEqual(self.sq.build_query(), "(hello AND world)")

        self.sq.add_filter(SQ(content__exact="hello world"))
        self.assertEqual(
            self.sq.build_query(), '((hello AND world) AND ("hello world"))'
        )

    def test_build_query_boost(self):
        self.sq.add_filter(SQ(content="hello"))
        self.sq.add_boost("world", 5)
        self.assertEqual(self.sq.build_query(), "(hello) world^5")

    def test_build_query_multiple_filter_types(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(pub_date__lte=Exact("2009-02-10 01:59:00")))
        self.sq.add_filter(SQ(author__gt="daniel"))
        self.sq.add_filter(SQ(created__lt=Exact("2009-02-12 12:13:00")))
        self.sq.add_filter(SQ(title__gte="B"))
        self.sq.add_filter(SQ(id__in=[1, 2, 3]))
        self.sq.add_filter(SQ(rating__range=[3, 5]))
        self.assertEqual(
            self.sq.build_query(),
            '((why) AND pub_date:([* TO "2009-02-10 01:59:00"]) AND author:({"daniel" TO *}) AND created:({* TO "2009-02-12 12:13:00"}) AND title:(["B" TO *]) AND id:("1" OR "2" OR "3") AND rating:(["3" TO "5"]))',
        )

    def test_build_query_multiple_filter_types_with_datetimes(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(pub_date__lte=datetime.datetime(2009, 2, 10, 1, 59, 0)))
        self.sq.add_filter(SQ(author__gt="daniel"))
        self.sq.add_filter(SQ(created__lt=datetime.datetime(2009, 2, 12, 12, 13, 0)))
        self.sq.add_filter(SQ(title__gte="B"))
        self.sq.add_filter(SQ(id__in=[1, 2, 3]))
        self.sq.add_filter(SQ(rating__range=[3, 5]))
        self.assertEqual(
            self.sq.build_query(),
            '((why) AND pub_date:([* TO "2009-02-10T01:59:00"]) AND author:({"daniel" TO *}) AND created:({* TO "2009-02-12T12:13:00"}) AND title:(["B" TO *]) AND id:("1" OR "2" OR "3") AND rating:(["3" TO "5"]))',
        )

    def test_build_query_in_filter_multiple_words(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(title__in=["A Famous Paper", "An Infamous Article"]))
        self.assertEqual(
            self.sq.build_query(),
            '((why) AND title:("A Famous Paper" OR "An Infamous Article"))',
        )

    def test_build_query_in_filter_datetime(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(pub_date__in=[datetime.datetime(2009, 7, 6, 1, 56, 21)]))
        self.assertEqual(
            self.sq.build_query(), '((why) AND pub_date:("2009-07-06T01:56:21"))'
        )

    def test_build_query_in_with_set(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(title__in={"A Famous Paper", "An Infamous Article"}))
        self.assertTrue("((why) AND title:(" in self.sq.build_query())
        self.assertTrue('"A Famous Paper"' in self.sq.build_query())
        self.assertTrue('"An Infamous Article"' in self.sq.build_query())

    def test_build_query_wildcard_filter_types(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(title__startswith="haystack"))
        self.assertEqual(self.sq.build_query(), "((why) AND title:(haystack*))")

    def test_build_query_fuzzy_filter_types(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(title__fuzzy="haystack"))
        self.assertEqual(self.sq.build_query(), "((why) AND title:(haystack~))")

    def test_clean(self):
        self.assertEqual(self.sq.clean("hello world"), "hello world")
        self.assertEqual(self.sq.clean("hello AND world"), "hello and world")
        self.assertEqual(
            self.sq.clean(
                r'hello AND OR NOT TO + - && || ! ( ) { } [ ] ^ " ~ * ? : \ / world'
            ),
            'hello and or not to \\+ \\- \\&& \\|| \\! \\( \\) \\{ \\} \\[ \\] \\^ \\" \\~ \\* \\? \\: \\\\ \\/ world',
        )
        self.assertEqual(
            self.sq.clean("so please NOTe i am in a bAND and bORed"),
            "so please NOTe i am in a bAND and bORed",
        )

    def test_build_query_with_models(self):
        self.sq.add_filter(SQ(content="hello"))
        self.sq.add_model(MockModel)
        self.assertEqual(self.sq.build_query(), "(hello)")

        self.sq.add_model(AnotherMockModel)
        self.assertEqual(self.sq.build_query(), "(hello)")

    def test_set_result_class(self):
        # Assert that we're defaulting to ``SearchResult``.
        self.assertTrue(issubclass(self.sq.result_class, SearchResult))

        # Custom class.
        class IttyBittyResult:
            pass

        self.sq.set_result_class(IttyBittyResult)
        self.assertTrue(issubclass(self.sq.result_class, IttyBittyResult))

        # Reset to default.
        self.sq.set_result_class(None)
        self.assertTrue(issubclass(self.sq.result_class, SearchResult))

    def test_in_filter_values_list(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(title__in=[1, 2, 3]))
        self.assertEqual(self.sq.build_query(), '((why) AND title:("1" OR "2" OR "3"))')

    def test_narrow_sq(self):
        sqs = SearchQuerySet(using="elasticsearch").narrow(SQ(foo="moof"))
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.narrow_queries), 1)
        self.assertEqual(sqs.query.narrow_queries.pop(), "foo:(moof)")

    def test_build_query_with_dwithin_range(self):
        from django.contrib.gis.geos import Point

        backend = connections["elasticsearch"].get_backend()
        search_kwargs = backend.build_search_kwargs(
            "where",
            dwithin={
                "field": "location_field",
                "point": Point(1.2345678, 2.3456789),
                "distance": D(m=500),
            },
        )
        self.assertEqual(
            search_kwargs["query"]["bool"]["filter"]["bool"]["must"][1]["geo_distance"],
            {
                "distance": "0.500000km",
                "location_field": {"lat": 2.3456789, "lon": 1.2345678},
            },
        )



from django.conf import settings
from django.contrib.auth.models import User
from django.test import TestCase
from django.test.utils import override_settings
from django.urls import reverse

from haystack import connections, reset_search_queries
from haystack.utils.loading import UnifiedIndex

from ..core.models import MockModel
from .test_solr_backend import SolrMockModelSearchIndex, clear_solr_index


@override_settings(DEBUG=True)
class SearchModelAdminTestCase(TestCase):
    fixtures = ["base_data.json", "bulk_data.json"]

    def setUp(self):
        super().setUp()

        # With the models setup, you get the proper bits.
        # Stow.
        self.old_ui = connections["solr"].get_unified_index()
        self.ui = UnifiedIndex()
        smmsi = SolrMockModelSearchIndex()
        self.ui.build(indexes=[smmsi])
        connections["solr"]._index = self.ui

        # Wipe it clean.
        clear_solr_index()

        # Force indexing of the content.
        smmsi.update(using="solr")

        superuser = User.objects.create_superuser(
            username="superuser", password="password", email="super@user.com"
        )

    def tearDown(self):
        # Restore.
        connections["solr"]._index = self.old_ui
        super().tearDown()

    def test_usage(self):
        reset_search_queries()
        self.assertEqual(len(connections["solr"].queries), 0)

        self.assertEqual(
            self.client.login(username="superuser", password="password"), True
        )

        # First, non-search behavior.
        resp = self.client.get("/admin/core/mockmodel/")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(connections["solr"].queries), 0)
        self.assertEqual(resp.context["cl"].full_result_count, 23)

        # Then search behavior.
        resp = self.client.get("/admin/core/mockmodel/", data={"q": "Haystack"})
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(connections["solr"].queries), 3)
        self.assertEqual(resp.context["cl"].full_result_count, 23)
        # Ensure they aren't search results.
        self.assertEqual(isinstance(resp.context["cl"].result_list[0], MockModel), True)

        result_pks = [i.pk for i in resp.context["cl"].result_list]
        self.assertIn(5, result_pks)

        # Make sure only changelist is affected.
        resp = self.client.get(reverse("admin:core_mockmodel_change", args=(1,)))
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.context["original"].id, 1)
        self.assertTemplateUsed(resp, "admin/change_form.html")

        # The Solr query count should be unchanged:
        self.assertEqual(len(connections["solr"].queries), 3)



import datetime

from django.db import models


class Cat(models.Model):
    name = models.CharField(max_length=255)
    birth_date = models.DateField(default=datetime.date.today)
    bio = models.TextField(blank=True)
    created = models.DateTimeField(default=datetime.datetime.now)
    updated = models.DateTimeField(default=datetime.datetime.now)

    def __str__(self):
        return self.name

    @models.permalink
    def get_absolute_url(self):
        return ("cat_detail", [], {"id": self.id})



import os
from tempfile import mkdtemp

SECRET_KEY = "Please do not spew DeprecationWarnings"

# Haystack settings for running tests.
DATABASES = {
    "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": "haystack_tests.db"}
}

# Use BigAutoField as the default auto field for all models
DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "haystack",
    "test_haystack.discovery",
    "test_haystack.core",
    "test_haystack.spatial",
    "test_haystack.multipleindex",
    # This app exists to confirm that nothing breaks when INSTALLED_APPS has an app without models.py
    # which is common in some cases for things like admin extensions, reporting, etc.
    "test_haystack.test_app_without_models",
    # Confirm that everything works with app labels which have more than one level of hierarchy
    # as reported in https://github.com/django-haystack/django-haystack/issues/1152
    "test_haystack.test_app_with_hierarchy.contrib.django.hierarchal_app_django",
    "test_haystack.test_app_using_appconfig.apps.SimpleTestAppConfig",
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]

MIDDLEWARE = [
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]

ROOT_URLCONF = "test_haystack.core.urls"

HAYSTACK_ROUTERS = [
    "haystack.routers.DefaultRouter",
    "test_haystack.multipleindex.routers.MultipleIndexRouter",
]

HAYSTACK_CONNECTIONS = {
    "default": {"ENGINE": "test_haystack.mocks.MockEngine"},
    "whoosh": {
        "ENGINE": "haystack.backends.whoosh_backend.WhooshEngine",
        "PATH": mkdtemp(prefix="test_whoosh_query"),
        "INCLUDE_SPELLING": True,
    },
    "filtered_whoosh": {
        "ENGINE": "haystack.backends.whoosh_backend.WhooshEngine",
        "PATH": mkdtemp(prefix="haystack-multipleindex-filtered-whoosh-tests-"),
        "EXCLUDED_INDEXES": ["test_haystack.multipleindex.search_indexes.BarIndex"],
    },
    "elasticsearch": {
        "ENGINE": "haystack.backends.elasticsearch_backend.ElasticsearchSearchEngine",
        "URL": os.environ.get("TEST_ELASTICSEARCH_1_URL", "http://localhost:9200/"),
        "INDEX_NAME": "test_default",
        "INCLUDE_SPELLING": True,
    },
    "simple": {"ENGINE": "haystack.backends.simple_backend.SimpleEngine"},
    "solr": {
        "ENGINE": "haystack.backends.solr_backend.SolrEngine",
        "URL": os.environ.get(
            "TEST_SOLR_URL", "http://localhost:9001/solr/collection1"
        ),
        "ADMIN_URL": os.environ.get(
            "TEST_SOLR_ADMIN_URL", "http://localhost:9001/solr/admin/cores"
        ),
        "INCLUDE_SPELLING": True,
    },
}

if "elasticsearch" in HAYSTACK_CONNECTIONS:
    try:
        import elasticsearch

        if (2,) <= elasticsearch.__version__ <= (3,):
            HAYSTACK_CONNECTIONS["elasticsearch"].update(
                {
                    "ENGINE": "haystack.backends.elasticsearch2_backend.Elasticsearch2SearchEngine"
                }
            )
        elif (5,) <= elasticsearch.__version__ <= (6,):
            HAYSTACK_CONNECTIONS["elasticsearch"].update(
                {
                    "ENGINE": "haystack.backends.elasticsearch5_backend.Elasticsearch5SearchEngine"
                }
            )
        elif (7,) <= elasticsearch.__version__ <= (8,):
            HAYSTACK_CONNECTIONS["elasticsearch"].update(
                {
                    "ENGINE": "haystack.backends.elasticsearch7_backend.Elasticsearch7SearchEngine"
                }
            )
    except ImportError:
        del HAYSTACK_CONNECTIONS["elasticsearch"]



import datetime

from django.test import TestCase

from haystack import connections
from haystack.inputs import AltParser, Exact
from haystack.models import SearchResult
from haystack.query import SQ, SearchQuerySet

from ..core.models import AnotherMockModel, MockModel


class SolrSearchQueryTestCase(TestCase):
    fixtures = ["base_data"]

    def setUp(self):
        super().setUp()
        self.sq = connections["solr"].get_query()

    def test_build_query_all(self):
        self.assertEqual(self.sq.build_query(), "*:*")

    def test_build_query_single_word(self):
        self.sq.add_filter(SQ(content="hello"))
        self.assertEqual(self.sq.build_query(), "(hello)")

    def test_build_query_boolean(self):
        self.sq.add_filter(SQ(content=True))
        self.assertEqual(self.sq.build_query(), "(true)")

    def test_build_query_datetime(self):
        self.sq.add_filter(SQ(content=datetime.datetime(2009, 5, 8, 11, 28)))
        self.assertEqual(self.sq.build_query(), "(2009-05-08T11:28:00Z)")

    def test_build_query_multiple_words_and(self):
        self.sq.add_filter(SQ(content="hello"))
        self.sq.add_filter(SQ(content="world"))
        self.assertEqual(self.sq.build_query(), "((hello) AND (world))")

    def test_build_query_multiple_words_not(self):
        self.sq.add_filter(~SQ(content="hello"))
        self.sq.add_filter(~SQ(content="world"))
        self.assertEqual(self.sq.build_query(), "(NOT ((hello)) AND NOT ((world)))")

    def test_build_query_multiple_words_or(self):
        self.sq.add_filter(~SQ(content="hello"))
        self.sq.add_filter(SQ(content="hello"), use_or=True)
        self.assertEqual(self.sq.build_query(), "(NOT ((hello)) OR (hello))")

    def test_build_query_multiple_words_mixed(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(content="hello"), use_or=True)
        self.sq.add_filter(~SQ(content="world"))
        self.assertEqual(
            self.sq.build_query(), "(((why) OR (hello)) AND NOT ((world)))"
        )

    def test_build_query_phrase(self):
        self.sq.add_filter(SQ(content="hello world"))
        self.assertEqual(self.sq.build_query(), "(hello AND world)")

        self.sq.add_filter(SQ(content__exact="hello world"))
        self.assertEqual(
            self.sq.build_query(), '((hello AND world) AND ("hello world"))'
        )

    def test_build_query_boost(self):
        self.sq.add_filter(SQ(content="hello"))
        self.sq.add_boost("world", 5)
        self.assertEqual(self.sq.build_query(), "(hello) world^5")

    def test_correct_exact(self):
        self.sq.add_filter(SQ(content=Exact("hello world")))
        self.assertEqual(self.sq.build_query(), '("hello world")')

    def test_build_query_multiple_filter_types(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(pub_date__lte=Exact("2009-02-10 01:59:00")))
        self.sq.add_filter(SQ(author__gt="daniel"))
        self.sq.add_filter(SQ(created__lt=Exact("2009-02-12 12:13:00")))
        self.sq.add_filter(SQ(title__gte="B"))
        self.sq.add_filter(SQ(id__in=[1, 2, 3]))
        self.sq.add_filter(SQ(rating__range=[3, 5]))
        self.assertEqual(
            self.sq.build_query(),
            '((why) AND pub_date:([* TO "2009-02-10 01:59:00"]) AND author:({"daniel" TO *}) AND created:({* TO "2009-02-12 12:13:00"}) AND title:(["B" TO *]) AND id:("1" OR "2" OR "3") AND rating:(["3" TO "5"]))',
        )

    def test_build_complex_altparser_query(self):
        self.sq.add_filter(SQ(content=AltParser("dismax", "Don't panic", qf="text")))
        self.sq.add_filter(SQ(pub_date__lte=Exact("2009-02-10 01:59:00")))
        self.sq.add_filter(SQ(author__gt="daniel"))
        self.sq.add_filter(SQ(created__lt=Exact("2009-02-12 12:13:00")))
        self.sq.add_filter(SQ(title__gte="B"))
        self.sq.add_filter(SQ(id__in=[1, 2, 3]))
        self.sq.add_filter(SQ(rating__range=[3, 5]))
        query = self.sq.build_query()
        self.assertTrue('(_query_:"{!dismax qf=text}Don\'t panic")' in query)
        self.assertTrue('pub_date:([* TO "2009-02-10 01:59:00"])' in query)
        self.assertTrue('author:({"daniel" TO *})' in query)
        self.assertTrue('created:({* TO "2009-02-12 12:13:00"})' in query)
        self.assertTrue('title:(["B" TO *])' in query)
        self.assertTrue('id:("1" OR "2" OR "3")' in query)
        self.assertTrue('rating:(["3" TO "5"])' in query)

    def test_build_query_multiple_filter_types_with_datetimes(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(pub_date__lte=datetime.datetime(2009, 2, 10, 1, 59, 0)))
        self.sq.add_filter(SQ(author__gt="daniel"))
        self.sq.add_filter(SQ(created__lt=datetime.datetime(2009, 2, 12, 12, 13, 0)))
        self.sq.add_filter(SQ(title__gte="B"))
        self.sq.add_filter(SQ(id__in=[1, 2, 3]))
        self.sq.add_filter(SQ(rating__range=[3, 5]))
        self.assertEqual(
            self.sq.build_query(),
            '((why) AND pub_date:([* TO "2009-02-10T01:59:00Z"]) AND author:({"daniel" TO *}) AND created:({* TO "2009-02-12T12:13:00Z"}) AND title:(["B" TO *]) AND id:("1" OR "2" OR "3") AND rating:(["3" TO "5"]))',
        )

    def test_build_query_in_filter_multiple_words(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(title__in=["A Famous Paper", "An Infamous Article"]))
        self.assertEqual(
            self.sq.build_query(),
            '((why) AND title:("A Famous Paper" OR "An Infamous Article"))',
        )

    def test_build_query_in_filter_datetime(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(pub_date__in=[datetime.datetime(2009, 7, 6, 1, 56, 21)]))
        self.assertEqual(
            self.sq.build_query(), '((why) AND pub_date:("2009-07-06T01:56:21Z"))'
        )

    def test_build_query_in_with_set(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(title__in=set(["A Famous Paper", "An Infamous Article"])))
        query = self.sq.build_query()
        self.assertTrue("(why)" in query)

        # Because ordering in Py3 is now random.
        if 'title:("A ' in query:
            self.assertTrue(
                'title:("A Famous Paper" OR "An Infamous Article")' in query
            )
        else:
            self.assertTrue(
                'title:("An Infamous Article" OR "A Famous Paper")' in query
            )

    def test_build_query_with_contains(self):
        self.sq.add_filter(SQ(content="circular"))
        self.sq.add_filter(SQ(title__contains="haystack"))
        self.assertEqual(self.sq.build_query(), "((circular) AND title:(*haystack*))")

    def test_build_query_with_endswith(self):
        self.sq.add_filter(SQ(content="circular"))
        self.sq.add_filter(SQ(title__endswith="haystack"))
        self.assertEqual(self.sq.build_query(), "((circular) AND title:(*haystack))")

    def test_build_query_wildcard_filter_types(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(title__startswith="haystack"))
        self.assertEqual(self.sq.build_query(), "((why) AND title:(haystack*))")

    def test_build_query_fuzzy_filter_types(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(title__fuzzy="haystack"))
        self.assertEqual(self.sq.build_query(), "((why) AND title:(haystack~))")

    def test_clean(self):
        self.assertEqual(self.sq.clean("hello world"), "hello world")
        self.assertEqual(self.sq.clean("hello AND world"), "hello and world")
        self.assertEqual(
            self.sq.clean(
                'hello AND OR NOT TO + - && || ! ( ) { } [ ] ^ " ~ * ? : \ / world'
            ),
            'hello and or not to \\+ \\- \\&& \\|| \\! \\( \\) \\{ \\} \\[ \\] \\^ \\" \\~ \\* \\? \\: \\\\ \\/ world',
        )
        self.assertEqual(
            self.sq.clean("so please NOTe i am in a bAND and bORed"),
            "so please NOTe i am in a bAND and bORed",
        )

    def test_build_query_with_models(self):
        self.sq.add_filter(SQ(content="hello"))
        self.sq.add_model(MockModel)
        self.assertEqual(self.sq.build_query(), "(hello)")

        self.sq.add_model(AnotherMockModel)
        self.assertEqual(self.sq.build_query(), "(hello)")

    def test_set_result_class(self):
        # Assert that we're defaulting to ``SearchResult``.
        self.assertTrue(issubclass(self.sq.result_class, SearchResult))

        # Custom class.
        class IttyBittyResult:
            pass

        self.sq.set_result_class(IttyBittyResult)
        self.assertTrue(issubclass(self.sq.result_class, IttyBittyResult))

        # Reset to default.
        self.sq.set_result_class(None)
        self.assertTrue(issubclass(self.sq.result_class, SearchResult))

    def test_in_filter_values_list(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(title__in=MockModel.objects.values_list("id", flat=True)))
        self.assertEqual(self.sq.build_query(), '((why) AND title:("1" OR "2" OR "3"))')

    def test_narrow_sq(self):
        sqs = SearchQuerySet(using="solr").narrow(SQ(foo="moof"))
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.narrow_queries), 1)
        self.assertEqual(sqs.query.narrow_queries.pop(), "foo:(moof)")

    def test_query__in(self):
        sqs = SearchQuerySet(using="solr").filter(id__in=[1, 2, 3])
        self.assertEqual(sqs.query.build_query(), 'id:("1" OR "2" OR "3")')

    def test_query__in_empty_list(self):
        """Confirm that an empty list avoids a Solr exception"""
        sqs = SearchQuerySet(using="solr").filter(id__in=[])
        self.assertEqual(sqs.query.build_query(), "id:(!*:*)")



import datetime

from django.db import models


class Checkin(models.Model):
    username = models.CharField(max_length=255)
    # We're going to do some non-GeoDjango action, since the setup is
    # complex enough. You could just as easily do:
    #
    #   location = models.PointField()
    #
    # ...and your ``search_indexes.py`` could be less complex.
    latitude = models.FloatField()
    longitude = models.FloatField()
    comment = models.CharField(
        max_length=140, blank=True, default="", help_text="Say something pithy."
    )
    created = models.DateTimeField(default=datetime.datetime.now)

    class Meta:
        ordering = ["-created"]

    # Again, with GeoDjango, this would be unnecessary.
    def get_location(self):
        # Nothing special about this Point, but ensure that's we don't have to worry
        # about import paths.
        from django.contrib.gis.geos import Point

        pnt = Point(self.longitude, self.latitude)
        return pnt



from django.conf import settings
from django.test import TestCase

from haystack import connection_router, connections, constants, indexes
from haystack.management.commands.build_solr_schema import Command
from haystack.query import SQ
from haystack.utils.loading import UnifiedIndex
from test_haystack.core.models import AnotherMockModel, MockModel
from test_haystack.utils import check_solr


class MockModelSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(model_attr="foo", document=True)
    name = indexes.CharField(model_attr="author")
    pub_date = indexes.DateTimeField(model_attr="pub_date")

    def get_model(self):
        return MockModel


class AlteredInternalNamesTestCase(TestCase):
    def setUp(self):
        check_solr()
        super().setUp()

        self.old_ui = connections["solr"].get_unified_index()
        ui = UnifiedIndex()
        ui.build(indexes=[MockModelSearchIndex()])
        connections["solr"]._index = ui

        constants.ID = "my_id"
        constants.DJANGO_CT = "my_django_ct"
        constants.DJANGO_ID = "my_django_id"

    def tearDown(self):
        constants.ID = "id"
        constants.DJANGO_CT = "django_ct"
        constants.DJANGO_ID = "django_id"
        connections["solr"]._index = self.old_ui
        super().tearDown()

    def test_altered_names(self):
        sq = connections["solr"].get_query()

        sq.add_filter(SQ(content="hello"))
        sq.add_model(MockModel)
        self.assertEqual(sq.build_query(), "(hello)")

        sq.add_model(AnotherMockModel)
        self.assertEqual(sq.build_query(), "(hello)")

    def test_solr_schema(self):
        command = Command()
        context_data = command.build_context(using="solr")
        self.assertEqual(len(context_data), 6)
        self.assertEqual(context_data["DJANGO_ID"], "my_django_id")
        self.assertEqual(context_data["content_field_name"], "text")
        self.assertEqual(context_data["DJANGO_CT"], "my_django_ct")
        self.assertEqual(context_data["default_operator"], "AND")
        self.assertEqual(context_data["ID"], "my_id")
        self.assertEqual(len(context_data["fields"]), 3)
        self.assertEqual(
            sorted(context_data["fields"], key=lambda x: x["field_name"]),
            [
                {
                    "indexed": "true",
                    "type": "text_en",
                    "stored": "true",
                    "field_name": "name",
                    "multi_valued": "false",
                },
                {
                    "indexed": "true",
                    "type": "date",
                    "stored": "true",
                    "field_name": "pub_date",
                    "multi_valued": "false",
                },
                {
                    "indexed": "true",
                    "type": "text_en",
                    "stored": "true",
                    "field_name": "text",
                    "multi_valued": "false",
                },
            ],
        )

        schema_xml = command.build_template(using="solr")
        self.assertTrue("<uniqueKey>my_id</uniqueKey>" in schema_xml)
        self.assertTrue(
            '<field name="my_id" type="string" indexed="true" stored="true" multiValued="false" required="true"/>'
            in schema_xml
        )
        self.assertTrue(
            '<field name="my_django_ct" type="string" indexed="true" stored="true" multiValued="false"/>'
            in schema_xml
        )



import warnings

from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase

from haystack.utils import loading


class LoadBackendTestCase(TestCase):
    def test_load_solr(self):
        try:
            import pysolr
        except ImportError:
            warnings.warn(
                "Pysolr doesn't appear to be installed. Unable to test loading the Solr backend."
            )
            return

        backend = loading.load_backend("haystack.backends.solr_backend.SolrEngine")
        self.assertEqual(backend.__name__, "SolrEngine")

    def test_load_whoosh(self):
        try:
            import whoosh
        except ImportError:
            warnings.warn(
                "Whoosh doesn't appear to be installed. Unable to test loading the Whoosh backend."
            )
            return

        backend = loading.load_backend("haystack.backends.whoosh_backend.WhooshEngine")
        self.assertEqual(backend.__name__, "WhooshEngine")

    def test_load_elasticsearch(self):
        try:
            import elasticsearch
        except ImportError:
            warnings.warn(
                "elasticsearch-py doesn't appear to be installed. Unable to test loading the ElasticSearch backend."
            )
            return

        backend = loading.load_backend(
            "haystack.backends.elasticsearch_backend.ElasticsearchSearchEngine"
        )
        self.assertEqual(backend.__name__, "ElasticsearchSearchEngine")

    def test_load_simple(self):
        backend = loading.load_backend("haystack.backends.simple_backend.SimpleEngine")
        self.assertEqual(backend.__name__, "SimpleEngine")

    def test_load_nonexistent(self):
        try:
            backend = loading.load_backend("foobar")
            self.fail()
        except ImproperlyConfigured as e:
            self.assertEqual(
                str(e),
                "The provided backend 'foobar' is not a complete Python path to a BaseEngine subclass.",
            )

        try:
            backend = loading.load_backend("foobar.FooEngine")
            self.fail()
        except ImportError as e:
            pass

        try:
            backend = loading.load_backend("haystack.backends.simple_backend.FooEngine")
            self.fail()
        except ImportError as e:
            self.assertEqual(
                str(e),
                "The Python module 'haystack.backends.simple_backend' has no 'FooEngine' class.",
            )



import unittest

from django.conf import settings


def check_solr(using="solr"):
    try:
        from pysolr import Solr, SolrError
    except ImportError:
        raise unittest.SkipTest("pysolr not installed.")

    solr = Solr(settings.HAYSTACK_CONNECTIONS[using]["URL"])
    try:
        solr.search("*:*")
    except SolrError as e:
        raise unittest.SkipTest(
            "solr not running on %r" % settings.HAYSTACK_CONNECTIONS[using]["URL"], e
        )



from django.contrib import admin
from django.urls import include, path

from haystack.forms import FacetedSearchForm
from haystack.query import SearchQuerySet
from haystack.views import FacetedSearchView, SearchView, basic_search

admin.autodiscover()


urlpatterns = [
    path("", SearchView(load_all=False), name="haystack_search"),
    path("admin/", admin.site.urls),
    path("basic/", basic_search, {"load_all": False}, name="haystack_basic_search"),
    path(
        "faceted/",
        FacetedSearchView(
            searchqueryset=SearchQuerySet().facet("author"),
            form_class=FacetedSearchForm,
        ),
        name="haystack_faceted_search",
    ),
]

urlpatterns += [
    path(
        "",
        include(("test_haystack.test_app_without_models.urls", "app-without-models")),
    )
]



from django.test import TestCase
from django.test.utils import override_settings

from haystack.utils import (
    _lookup_identifier_method,
    get_facet_field_name,
    get_identifier,
    log,
)
from haystack.utils.highlighting import Highlighter
from test_haystack.core.models import MockModel


class GetIdentifierTestCase(TestCase):
    def test_get_facet_field_name(self):
        self.assertEqual(get_facet_field_name("id"), "id")
        self.assertEqual(get_facet_field_name("django_id"), "django_id")
        self.assertEqual(get_facet_field_name("django_ct"), "django_ct")
        self.assertEqual(get_facet_field_name("author"), "author_exact")
        self.assertEqual(get_facet_field_name("author_exact"), "author_exact_exact")


class GetFacetFieldNameTestCase(TestCase):
    fixtures = ["base_data"]

    def test_get_identifier(self):
        self.assertEqual(get_identifier("core.mockmodel.1"), "core.mockmodel.1")

        # Valid object.
        mock = MockModel.objects.get(pk=1)
        self.assertEqual(get_identifier(mock), "core.mockmodel.1")

    @override_settings(
        HAYSTACK_IDENTIFIER_METHOD="test_haystack.core.custom_identifier.get_identifier_method"
    )
    def test_haystack_identifier_method(self):
        # The custom implementation returns the MD-5 hash of the key value by
        # default:
        get_identifier = _lookup_identifier_method()
        self.assertEqual(get_identifier("a.b.c"), "553f764f7b436175c0387e22b4a19213")

        # … but it also supports a custom override mechanism which would
        # definitely fail with the default implementation:
        class custom_id_class:
            def get_custom_haystack_id(self):
                return "CUSTOM"

        self.assertEqual(get_identifier(custom_id_class()), "CUSTOM")

    @override_settings(
        HAYSTACK_IDENTIFIER_METHOD="test_haystack.core.custom_identifier.not_there"
    )
    def test_haystack_identifier_method_bad_path(self):
        self.assertRaises(AttributeError, _lookup_identifier_method)

    @override_settings(HAYSTACK_IDENTIFIER_METHOD="core.not_there.not_there")
    def test_haystack_identifier_method_bad_module(self):
        self.assertRaises(ImportError, _lookup_identifier_method)


class HighlighterTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.document_1 = "This is a test of the highlightable words detection. This is only a test. Were this an actual emergency, your text would have exploded in mid-air."
        self.document_2 = (
            "The content of words in no particular order causes nothing to occur."
        )
        self.document_3 = "%s %s" % (self.document_1, self.document_2)

    def test_find_highlightable_words(self):
        highlighter = Highlighter("this test")
        highlighter.text_block = self.document_1
        self.assertEqual(
            highlighter.find_highlightable_words(),
            {"this": [0, 53, 79], "test": [10, 68]},
        )

        # We don't stem for now.
        highlighter = Highlighter("highlight tests")
        highlighter.text_block = self.document_1
        self.assertEqual(
            highlighter.find_highlightable_words(), {"highlight": [22], "tests": []}
        )

        # Ignore negated bits.
        highlighter = Highlighter("highlight -test")
        highlighter.text_block = self.document_1
        self.assertEqual(highlighter.find_highlightable_words(), {"highlight": [22]})

    def test_find_window(self):
        # The query doesn't matter for this method, so ignore it.
        highlighter = Highlighter("")
        highlighter.text_block = self.document_1

        # No query.
        self.assertEqual(highlighter.find_window({}), (0, 200))

        # Nothing found.
        self.assertEqual(
            highlighter.find_window({"highlight": [], "tests": []}), (0, 200)
        )

        # Simple cases.
        self.assertEqual(
            highlighter.find_window({"highlight": [0], "tests": [100]}), (0, 200)
        )
        self.assertEqual(
            highlighter.find_window({"highlight": [99], "tests": [199]}), (99, 299)
        )
        self.assertEqual(
            highlighter.find_window({"highlight": [0], "tests": [201]}), (0, 200)
        )
        self.assertEqual(
            highlighter.find_window({"highlight": [203], "tests": [120]}), (120, 320)
        )
        self.assertEqual(
            highlighter.find_window({"highlight": [], "tests": [100]}), (100, 300)
        )
        self.assertEqual(
            highlighter.find_window({"highlight": [0], "tests": [80], "moof": [120]}),
            (0, 200),
        )

        # Simple cases, with an outlier far outside the window.
        self.assertEqual(
            highlighter.find_window({"highlight": [0], "tests": [100, 450]}), (0, 200)
        )
        self.assertEqual(
            highlighter.find_window({"highlight": [100], "tests": [220, 450]}),
            (100, 300),
        )
        self.assertEqual(
            highlighter.find_window({"highlight": [100], "tests": [350, 450]}),
            (350, 550),
        )
        self.assertEqual(
            highlighter.find_window(
                {"highlight": [100], "tests": [220], "moof": [450]}
            ),
            (100, 300),
        )

        # Density checks.
        self.assertEqual(
            highlighter.find_window({"highlight": [0], "tests": [100, 180, 450]}),
            (0, 200),
        )
        self.assertEqual(
            highlighter.find_window(
                {"highlight": [0, 40], "tests": [100, 200, 220, 450]}
            ),
            (40, 240),
        )
        self.assertEqual(
            highlighter.find_window(
                {"highlight": [0, 40], "tests": [100, 200, 220], "moof": [450]}
            ),
            (40, 240),
        )
        self.assertEqual(
            highlighter.find_window(
                {
                    "highlight": [0, 40],
                    "tests": [100, 200, 220],
                    "moof": [294, 299, 450],
                }
            ),
            (100, 300),
        )

    def test_render_html(self):
        highlighter = Highlighter("this test")
        highlighter.text_block = self.document_1
        self.assertEqual(
            highlighter.render_html({"this": [0, 53, 79], "test": [10, 68]}, 0, 200),
            '<span class="highlighted">This</span> is a <span class="highlighted">test</span> of the highlightable words detection. <span class="highlighted">This</span> is only a <span class="highlighted">test</span>. Were <span class="highlighted">this</span> an actual emergency, your text would have exploded in mid-air.',
        )

        highlighter.text_block = self.document_2
        self.assertEqual(
            highlighter.render_html({"this": [0, 53, 79], "test": [10, 68]}, 0, 200),
            "The content of words in no particular order causes nothing to occur.",
        )

        highlighter.text_block = self.document_3
        self.assertEqual(
            highlighter.render_html({"this": [0, 53, 79], "test": [10, 68]}, 0, 200),
            '<span class="highlighted">This</span> is a <span class="highlighted">test</span> of the highlightable words detection. <span class="highlighted">This</span> is only a <span class="highlighted">test</span>. Were <span class="highlighted">this</span> an actual emergency, your text would have exploded in mid-air. The content of words in no particular order causes no...',
        )

        highlighter = Highlighter("content detection")
        highlighter.text_block = self.document_3
        self.assertEqual(
            highlighter.render_html({"content": [151], "detection": [42]}, 42, 242),
            '...<span class="highlighted">detection</span>. This is only a test. Were this an actual emergency, your text would have exploded in mid-air. The <span class="highlighted">content</span> of words in no particular order causes nothing to occur.',
        )

        self.assertEqual(
            highlighter.render_html({"content": [151], "detection": [42]}, 42, 200),
            '...<span class="highlighted">detection</span>. This is only a test. Were this an actual emergency, your text would have exploded in mid-air. The <span class="highlighted">content</span> of words in no particular order causes no...',
        )

        # One term found within another term.
        highlighter = Highlighter("this is")
        highlighter.text_block = self.document_1
        self.assertEqual(
            highlighter.render_html(
                {"this": [0, 53, 79], "is": [2, 5, 55, 58, 81]}, 0, 200
            ),
            '<span class="highlighted">This</span> <span class="highlighted">is</span> a test of the highlightable words detection. <span class="highlighted">This</span> <span class="highlighted">is</span> only a test. Were <span class="highlighted">this</span> an actual emergency, your text would have exploded in mid-air.',
        )

        # Regression for repetition in the regular expression.
        highlighter = Highlighter("i++")
        highlighter.text_block = "Foo is i++ in most cases."
        self.assertEqual(
            highlighter.render_html({"i++": [7]}, 0, 200),
            'Foo is <span class="highlighted">i++</span> in most cases.',
        )
        highlighter = Highlighter("i**")
        highlighter.text_block = "Foo is i** in most cases."
        self.assertEqual(
            highlighter.render_html({"i**": [7]}, 0, 200),
            'Foo is <span class="highlighted">i**</span> in most cases.',
        )
        highlighter = Highlighter("i..")
        highlighter.text_block = "Foo is i.. in most cases."
        self.assertEqual(
            highlighter.render_html({"i..": [7]}, 0, 200),
            'Foo is <span class="highlighted">i..</span> in most cases.',
        )
        highlighter = Highlighter("i??")
        highlighter.text_block = "Foo is i?? in most cases."
        self.assertEqual(
            highlighter.render_html({"i??": [7]}, 0, 200),
            'Foo is <span class="highlighted">i??</span> in most cases.',
        )

        # Regression for highlighting already highlighted HTML terms.
        highlighter = Highlighter("span")
        highlighter.text_block = "A span in spam makes html in a can."
        self.assertEqual(
            highlighter.render_html({"span": [2]}, 0, 200),
            'A <span class="highlighted">span</span> in spam makes html in a can.',
        )

        highlighter = Highlighter("highlight")
        highlighter.text_block = "A span in spam makes highlighted html in a can."
        self.assertEqual(
            highlighter.render_html({"highlight": [21]}, 0, 200),
            'A span in spam makes <span class="highlighted">highlight</span>ed html in a can.',
        )

    def test_highlight(self):
        highlighter = Highlighter("this test")
        self.assertEqual(
            highlighter.highlight(self.document_1),
            '<span class="highlighted">This</span> is a <span class="highlighted">test</span> of the highlightable words detection. <span class="highlighted">This</span> is only a <span class="highlighted">test</span>. Were <span class="highlighted">this</span> an actual emergency, your text would have exploded in mid-air.',
        )
        self.assertEqual(
            highlighter.highlight(self.document_2),
            "The content of words in no particular order causes nothing to occur.",
        )
        self.assertEqual(
            highlighter.highlight(self.document_3),
            '<span class="highlighted">This</span> is a <span class="highlighted">test</span> of the highlightable words detection. <span class="highlighted">This</span> is only a <span class="highlighted">test</span>. Were <span class="highlighted">this</span> an actual emergency, your text would have exploded in mid-air. The content of words in no particular order causes no...',
        )

        highlighter = Highlighter("this test", html_tag="div", css_class=None)
        self.assertEqual(
            highlighter.highlight(self.document_1),
            "<div>This</div> is a <div>test</div> of the highlightable words detection. <div>This</div> is only a <div>test</div>. Were <div>this</div> an actual emergency, your text would have exploded in mid-air.",
        )
        self.assertEqual(
            highlighter.highlight(self.document_2),
            "The content of words in no particular order causes nothing to occur.",
        )
        self.assertEqual(
            highlighter.highlight(self.document_3),
            "<div>This</div> is a <div>test</div> of the highlightable words detection. <div>This</div> is only a <div>test</div>. Were <div>this</div> an actual emergency, your text would have exploded in mid-air. The content of words in no particular order causes no...",
        )

        highlighter = Highlighter("content detection")
        self.assertEqual(
            highlighter.highlight(self.document_1),
            '...<span class="highlighted">detection</span>. This is only a test. Were this an actual emergency, your text would have exploded in mid-air.',
        )
        self.assertEqual(
            highlighter.highlight(self.document_2),
            '...<span class="highlighted">content</span> of words in no particular order causes nothing to occur.',
        )
        self.assertEqual(
            highlighter.highlight(self.document_3),
            '...<span class="highlighted">detection</span>. This is only a test. Were this an actual emergency, your text would have exploded in mid-air. The <span class="highlighted">content</span> of words in no particular order causes nothing to occur.',
        )

        highlighter = Highlighter("content detection", max_length=100)
        self.assertEqual(
            highlighter.highlight(self.document_1),
            '...<span class="highlighted">detection</span>. This is only a test. Were this an actual emergency, your text would have exploded in mid-...',
        )
        self.assertEqual(
            highlighter.highlight(self.document_2),
            '...<span class="highlighted">content</span> of words in no particular order causes nothing to occur.',
        )
        self.assertEqual(
            highlighter.highlight(self.document_3),
            'This is a test of the highlightable words <span class="highlighted">detection</span>. This is only a test. Were this an actual emerge...',
        )


class LoggingFacadeTestCase(TestCase):
    def test_everything_noops_if_settings_are_off(self):
        with self.settings(HAYSTACK_LOGGING=False):
            l = log.LoggingFacade(None)
            l.error()

    def test_uses_provided_logger_if_logging_is_on(self):
        with self.settings(HAYSTACK_LOGGING=True):
            l = log.LoggingFacade(None)
            try:
                l.error()
            except AttributeError:
                pass

    def test_uses_provided_logger_by_default(self):
        class Logger:
            def __init__(self):
                self.was_called = False

            def error(self):
                self.was_called = True

        l = log.LoggingFacade(Logger())
        self.assertFalse(l.was_called, msg="sanity check")
        l.error()
        self.assertTrue(l.was_called)



import hashlib


def get_identifier_method(key):
    """
    Custom get_identifier method used for testing the
    setting HAYSTACK_IDENTIFIER_MODULE
    """

    if hasattr(key, "get_custom_haystack_id"):
        return key.get_custom_haystack_id()
    else:
        key_bytes = key.encode("utf-8")
        return hashlib.md5(key_bytes).hexdigest()



from datetime import date

from django.test import TestCase
from django.test.utils import override_settings

from haystack import connections
from haystack.query import SearchQuerySet
from haystack.utils.loading import UnifiedIndex

from ..core.models import MockModel, OneToManyRightSideModel, ScoreMockModel
from ..mocks import MockSearchResult
from .search_indexes import SimpleMockScoreIndex, SimpleMockSearchIndex


class SimpleSearchBackendTestCase(TestCase):
    fixtures = ["base_data.json", "bulk_data.json"]

    def setUp(self):
        super().setUp()

        self.backend = connections["simple"].get_backend()
        ui = connections["simple"].get_unified_index()
        self.index = SimpleMockSearchIndex()
        ui.build(indexes=[self.index, SimpleMockScoreIndex()])
        self.sample_objs = MockModel.objects.all()

    def test_update(self):
        self.backend.update(self.index, self.sample_objs)

    def test_remove(self):
        self.backend.remove(self.sample_objs[0])

    def test_clear(self):
        self.backend.clear()

    def test_search(self):
        # No query string should always yield zero results.
        self.assertEqual(self.backend.search(""), {"hits": 0, "results": []})

        self.assertEqual(self.backend.search("*")["hits"], 24)
        self.assertEqual(
            sorted([result.pk for result in self.backend.search("*")["results"]]),
            [
                1,
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                10,
                11,
                12,
                13,
                14,
                15,
                16,
                17,
                18,
                19,
                20,
                21,
                22,
                23,
            ],
        )

        self.assertEqual(self.backend.search("daniel")["hits"], 23)
        self.assertEqual(
            [result.pk for result in self.backend.search("daniel")["results"]],
            [
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                10,
                11,
                12,
                13,
                14,
                15,
                16,
                17,
                18,
                19,
                20,
                21,
                22,
                23,
            ],
        )

        self.assertEqual(self.backend.search("should be a string")["hits"], 1)
        self.assertEqual(
            [
                result.pk
                for result in self.backend.search("should be a string")["results"]
            ],
            [8],
        )
        # Ensure the results are ``SearchResult`` instances...
        self.assertEqual(
            self.backend.search("should be a string")["results"][0].score, 0
        )

        self.assertEqual(self.backend.search("index document")["hits"], 6)
        self.assertEqual(
            [result.pk for result in self.backend.search("index document")["results"]],
            [2, 3, 15, 16, 17, 18],
        )

        # Regression-ville
        self.assertEqual(
            [
                result.object.id
                for result in self.backend.search("index document")["results"]
            ],
            [2, 3, 15, 16, 17, 18],
        )
        self.assertEqual(
            self.backend.search("index document")["results"][0].model, MockModel
        )

        # No support for spelling suggestions
        self.assertEqual(self.backend.search("Indx")["hits"], 0)
        self.assertFalse(self.backend.search("Indx").get("spelling_suggestion"))

        # No support for facets
        self.assertEqual(
            self.backend.search("", facets=["name"]), {"hits": 0, "results": []}
        )
        self.assertEqual(self.backend.search("daniel", facets=["name"])["hits"], 23)
        self.assertEqual(
            self.backend.search(
                "",
                date_facets={
                    "pub_date": {
                        "start_date": date(2008, 2, 26),
                        "end_date": date(2008, 2, 26),
                        "gap": "/MONTH",
                    }
                },
            ),
            {"hits": 0, "results": []},
        )
        self.assertEqual(
            self.backend.search(
                "daniel",
                date_facets={
                    "pub_date": {
                        "start_date": date(2008, 2, 26),
                        "end_date": date(2008, 2, 26),
                        "gap": "/MONTH",
                    }
                },
            )["hits"],
            23,
        )
        self.assertEqual(
            self.backend.search("", query_facets={"name": "[* TO e]"}),
            {"hits": 0, "results": []},
        )
        self.assertEqual(
            self.backend.search("daniel", query_facets={"name": "[* TO e]"})["hits"], 23
        )
        self.assertFalse(self.backend.search("").get("facets"))
        self.assertFalse(self.backend.search("daniel").get("facets"))

        # Note that only textual-fields are supported.
        self.assertEqual(self.backend.search("2009-06-18")["hits"], 0)

        # Ensure that swapping the ``result_class`` works.
        self.assertTrue(
            isinstance(
                self.backend.search("index document", result_class=MockSearchResult)[
                    "results"
                ][0],
                MockSearchResult,
            )
        )

        # Ensure empty queries does not raise.
        self.assertEqual(
            self.backend.search("foo", models=[OneToManyRightSideModel]),
            {"hits": 0, "results": []},
        )

    def test_filter_models(self):
        self.backend.update(self.index, self.sample_objs)
        self.assertEqual(self.backend.search("*", models=set([]))["hits"], 24)
        self.assertEqual(self.backend.search("*", models=set([MockModel]))["hits"], 23)

    def test_more_like_this(self):
        self.backend.update(self.index, self.sample_objs)
        self.assertEqual(self.backend.search("*")["hits"], 24)

        # Unsupported by 'simple'. Should see empty results.
        self.assertEqual(self.backend.more_like_this(self.sample_objs[0])["hits"], 0)

    def test_score_field_collision(self):
        index = connections["simple"].get_unified_index().get_index(ScoreMockModel)
        sample_objs = ScoreMockModel.objects.all()

        self.backend.update(index, self.sample_objs)

        # 42 is the in the match, which will be removed from the result
        self.assertEqual(self.backend.search("42")["results"][0].score, 0)


@override_settings(DEBUG=True)
class LiveSimpleSearchQuerySetTestCase(TestCase):
    fixtures = ["base_data.json", "bulk_data.json"]

    def setUp(self):
        super().setUp()

        # Stow.
        self.old_ui = connections["simple"].get_unified_index()
        self.ui = UnifiedIndex()
        self.smmi = SimpleMockSearchIndex()
        self.ui.build(indexes=[self.smmi])
        connections["simple"]._index = self.ui

        self.sample_objs = MockModel.objects.all()
        self.sqs = SearchQuerySet(using="simple")

    def tearDown(self):
        # Restore.
        connections["simple"]._index = self.old_ui
        super().tearDown()

    def test_general_queries(self):
        # For now, just make sure these don't throw an exception.
        # They won't work until the simple backend is improved.
        self.assertTrue(len(self.sqs.auto_query("daniel")) > 0)
        self.assertTrue(len(self.sqs.filter(text="index")) > 0)
        self.assertTrue(len(self.sqs.exclude(name="daniel")) > 0)
        self.assertTrue(len(self.sqs.order_by("-pub_date")) > 0)

    def test_general_queries_unicode(self):
        self.assertEqual(len(self.sqs.auto_query("Привет")), 0)

    def test_more_like_this(self):
        # MLT shouldn't be horribly broken. This used to throw an exception.
        mm1 = MockModel.objects.get(pk=1)
        self.assertEqual(len(self.sqs.filter(text=1).more_like_this(mm1)), 0)

    def test_values_queries(self):
        sqs = self.sqs.auto_query("daniel")
        self.assertTrue(len(sqs) > 0)

        flat_scores = sqs.values_list("score", flat=True)
        self.assertEqual(flat_scores[0], 0)

        scores = sqs.values_list("id", "score")
        self.assertEqual(scores[0], [1, 0])

        scores_dict = sqs.values("id", "score")
        self.assertEqual(scores_dict[0], {"id": 1, "score": 0})



import datetime
import os
import shutil
import tempfile
from io import StringIO
from tempfile import mkdtemp
from unittest.mock import patch

import pysolr
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase

from haystack import connections, constants, indexes
from haystack.utils.loading import UnifiedIndex

from ..core.models import MockModel, MockTag


class SolrMockSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, use_template=True)
    name = indexes.CharField(model_attr="author", faceted=True)
    pub_date = indexes.DateTimeField(model_attr="pub_date")

    def get_model(self):
        return MockModel

    def get_updated_field(self):
        return "pub_date"


class SolrMockTagSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, model_attr="name")

    def get_model(self):
        return MockTag


class SolrMockSecretKeySearchIndex(indexes.SearchIndex, indexes.Indexable):
    Th3S3cr3tK3y = indexes.CharField(document=True, model_attr="author")

    def get_model(self):
        return MockModel


class ManagementCommandTestCase(TestCase):
    fixtures = ["base_data.json", "bulk_data.json"]

    def setUp(self):
        super().setUp()
        self.solr = pysolr.Solr(settings.HAYSTACK_CONNECTIONS["solr"]["URL"])

        # Stow.
        self.old_ui = connections["solr"].get_unified_index()
        self.ui = UnifiedIndex()
        self.smmi = SolrMockSearchIndex()
        self.ui.build(indexes=[self.smmi])
        connections["solr"]._index = self.ui

    def tearDown(self):
        connections["solr"]._index = self.old_ui
        super().tearDown()

    def verify_indexed_documents(self):
        """Confirm that the documents in the search index match the database"""

        res = self.solr.search("*:*", fl=["id"], rows=50)
        self.assertEqual(res.hits, 23)

        indexed_doc_ids = set(i["id"] for i in res.docs)
        expected_doc_ids = set(
            "core.mockmodel.%d" % i
            for i in MockModel.objects.values_list("pk", flat=True)
        )

        self.assertSetEqual(indexed_doc_ids, expected_doc_ids)

    def test_basic_commands(self):
        call_command("clear_index", interactive=False, verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        call_command("update_index", verbosity=0, commit=False)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        call_command("update_index", verbosity=0)
        self.verify_indexed_documents()

        call_command("clear_index", interactive=False, verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        call_command("rebuild_index", interactive=False, verbosity=0, commit=False)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        call_command("rebuild_index", interactive=False, verbosity=0, commit=True)
        self.verify_indexed_documents()

        call_command("clear_index", interactive=False, verbosity=0, commit=False)
        self.verify_indexed_documents()

    def test_remove(self):
        call_command("clear_index", interactive=False, verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        call_command("update_index", verbosity=0)
        self.verify_indexed_documents()

        # Remove several instances, two of which will fit in the same block:
        MockModel.objects.get(pk=1).delete()
        MockModel.objects.get(pk=2).delete()
        MockModel.objects.get(pk=8).delete()
        self.assertEqual(self.solr.search("*:*").hits, 23)

        # Plain ``update_index`` doesn't fix it.
        call_command("update_index", verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 23)

        # Remove without commit also doesn't affect queries:
        call_command(
            "update_index", remove=True, verbosity=0, batchsize=2, commit=False
        )
        self.assertEqual(self.solr.search("*:*").hits, 23)

        # … but remove with commit does:
        call_command("update_index", remove=True, verbosity=0, batchsize=2)
        self.assertEqual(self.solr.search("*:*").hits, 20)

    def test_age(self):
        call_command("clear_index", interactive=False, verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        start = datetime.datetime.now() - datetime.timedelta(hours=3)
        end = datetime.datetime.now()

        mock = MockModel.objects.get(pk=1)
        mock.pub_date = datetime.datetime.now() - datetime.timedelta(hours=2)
        mock.save()
        self.assertEqual(
            MockModel.objects.filter(pub_date__range=(start, end)).count(), 1
        )

        call_command("update_index", age=3, verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 1)

    def test_age_with_time_zones(self):
        """Haystack should use django.utils.timezone.now"""
        from django.utils.timezone import now as django_now

        from haystack.management.commands.update_index import now as haystack_now

        self.assertIs(
            haystack_now,
            django_now,
            msg="update_index should use django.utils.timezone.now",
        )

        with patch("haystack.management.commands.update_index.now") as m:
            m.return_value = django_now()
            self.test_age()
            assert m.called

    def test_dates(self):
        call_command("clear_index", interactive=False, verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        start = datetime.datetime.now() - datetime.timedelta(hours=5, minutes=30)
        end = datetime.datetime.now() - datetime.timedelta(hours=2)

        mock_1 = MockModel.objects.get(pk=1)
        mock_1.pub_date = datetime.datetime.now() - datetime.timedelta(
            hours=5, minutes=1
        )
        mock_1.save()
        mock_2 = MockModel.objects.get(pk=2)
        mock_2.pub_date = datetime.datetime.now() - datetime.timedelta(hours=3)
        mock_2.save()
        mock_3 = MockModel.objects.get(pk=3)
        mock_3.pub_date = datetime.datetime.now() - datetime.timedelta(hours=1)
        mock_3.save()
        self.assertEqual(
            MockModel.objects.filter(pub_date__range=(start, end)).count(), 2
        )

        call_command(
            "update_index",
            start_date=start.isoformat(),
            end_date=end.isoformat(),
            verbosity=0,
        )
        self.assertEqual(self.solr.search("*:*").hits, 2)

    def test_multiprocessing(self):
        call_command("clear_index", interactive=False, verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        call_command("update_index", verbosity=2, workers=2, batchsize=5)
        self.verify_indexed_documents()

        call_command("clear_index", interactive=False, verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        call_command("update_index", verbosity=2, workers=2, batchsize=5, commit=False)
        self.assertEqual(self.solr.search("*:*").hits, 0)

    def test_build_schema_wrong_backend(self):
        settings.HAYSTACK_CONNECTIONS["whoosh"] = {
            "ENGINE": "haystack.backends.whoosh_backend.WhooshEngine",
            "PATH": mkdtemp(prefix="dummy-path-"),
        }

        connections["whoosh"]._index = self.ui
        self.assertRaises(
            ImproperlyConfigured, call_command, "build_solr_schema", using="whoosh"
        )

    def test_build_schema(self):
        # Stow.
        oldhdf = constants.DOCUMENT_FIELD
        oldui = connections["solr"].get_unified_index()
        oldurl = settings.HAYSTACK_CONNECTIONS["solr"]["URL"]

        conf_dir = tempfile.mkdtemp()
        with open(os.path.join(conf_dir, "managed-schema"), "w+") as fp:
            pass
        try:
            needle = "Th3S3cr3tK3y"
            constants.DOCUMENT_FIELD = (
                needle  # Force index to use new key for document_fields
            )
            settings.HAYSTACK_CONNECTIONS["solr"]["URL"] = (
                settings.HAYSTACK_CONNECTIONS["solr"]["URL"].rsplit("/", 1)[0]
                + "/mgmnt"
            )

            ui = UnifiedIndex()
            ui.build(indexes=[SolrMockSecretKeySearchIndex()])
            connections["solr"]._index = ui

            rendered_file = StringIO()

            schema_file = os.path.join(conf_dir, "schema.xml")
            solrconfig_file = os.path.join(conf_dir, "solrconfig.xml")

            self.assertTrue(
                os.path.isdir(conf_dir), msg="Expected %s to be a directory" % conf_dir
            )

            call_command("build_solr_schema", using="solr", stdout=rendered_file)
            contents = rendered_file.getvalue()
            self.assertGreater(contents.find('name="%s' % needle), -1)

            call_command(
                "build_solr_schema", using="solr", configure_directory=conf_dir
            )
            with open(schema_file) as s:
                self.assertGreater(s.read().find('name="%s' % needle), -1)
            with open(solrconfig_file) as s:
                self.assertGreater(s.read().find('name="df">%s' % needle), -1)

            self.assertTrue(
                os.path.isfile(os.path.join(conf_dir, "managed-schema.old"))
            )

            with patch(
                "haystack.management.commands.build_solr_schema.requests.get"
            ) as mock_request:
                call_command("build_solr_schema", using="solr", reload_core=True)

            with patch(
                "haystack.management.commands.build_solr_schema.requests.get"
            ) as mock_request:
                mock_request.return_value.ok = False

                self.assertRaises(
                    CommandError,
                    call_command,
                    "build_solr_schema",
                    using="solr",
                    reload_core=True,
                )

            call_command("build_solr_schema", using="solr", filename=schema_file)
            with open(schema_file) as s:
                self.assertGreater(s.read().find('name="%s' % needle), -1)
        finally:
            # reset
            constants.DOCUMENT_FIELD = oldhdf
            connections["solr"]._index = oldui
            settings.HAYSTACK_CONNECTIONS["solr"]["URL"] = oldurl
            shutil.rmtree(conf_dir, ignore_errors=True)


class AppModelManagementCommandTestCase(TestCase):
    fixtures = ["base_data", "bulk_data.json"]

    def setUp(self):
        super().setUp()
        self.solr = pysolr.Solr(settings.HAYSTACK_CONNECTIONS["solr"]["URL"])

        # Stow.
        self.old_ui = connections["solr"].get_unified_index()
        self.ui = UnifiedIndex()
        self.smmi = SolrMockSearchIndex()
        self.smtmi = SolrMockTagSearchIndex()
        self.ui.build(indexes=[self.smmi, self.smtmi])
        connections["solr"]._index = self.ui

    def tearDown(self):
        connections["solr"]._index = self.old_ui
        super().tearDown()

    def test_app_model_variations(self):
        call_command("clear_index", interactive=False, verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        call_command("update_index", verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 25)

        call_command("clear_index", interactive=False, verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        call_command("update_index", "core", verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 25)

        call_command("clear_index", interactive=False, verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        with self.assertRaises(ImproperlyConfigured):
            call_command("update_index", "fake_app_thats_not_there")

        call_command("update_index", "core", "discovery", verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 25)

        call_command("clear_index", interactive=False, verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        call_command("update_index", "discovery", verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        call_command("clear_index", interactive=False, verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        call_command("update_index", "core.MockModel", verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 23)

        call_command("clear_index", interactive=False, verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        call_command("update_index", "core.MockTag", verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 2)

        call_command("clear_index", interactive=False, verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 0)

        call_command("update_index", "core.MockTag", "core.MockModel", verbosity=0)
        self.assertEqual(self.solr.search("*:*").hits, 25)



import os
import unittest

from django.conf import settings

from haystack.utils import log as logging


def load_tests(loader, standard_tests, pattern):
    log = logging.getLogger("haystack")
    try:
        import elasticsearch

        if not ((7, 0, 0) <= elasticsearch.__version__ < (8, 0, 0)):
            raise ImportError
        from elasticsearch import Elasticsearch, exceptions
    except ImportError:
        log.error(
            "Skipping ElasticSearch 7 tests: 'elasticsearch>=7.0.0,<8.0.0' not installed."
        )
        raise unittest.SkipTest("'elasticsearch>=7.0.0,<8.0.0' not installed.")

    url = settings.HAYSTACK_CONNECTIONS["elasticsearch"]["URL"]
    es = Elasticsearch(url)
    try:
        es.info()
    except exceptions.ConnectionError as e:
        log.error("elasticsearch not running on %r" % url, exc_info=True)
        raise unittest.SkipTest("elasticsearch not running on %r" % url, e)

    package_tests = loader.discover(
        start_dir=os.path.dirname(__file__), pattern=pattern
    )
    standard_tests.addTests(package_tests)
    return standard_tests



from types import GeneratorType, ModuleType

from django.test import TestCase
from django.urls import reverse

from haystack.utils import app_loading


class AppLoadingTests(TestCase):
    def test_load_apps(self):
        apps = app_loading.haystack_load_apps()
        self.assertIsInstance(apps, (list, GeneratorType))

        self.assertIn("hierarchal_app_django", apps)

        self.assertNotIn(
            "test_app_without_models",
            apps,
            msg="haystack_load_apps should exclude apps without defined models",
        )

    def test_get_app_modules(self):
        app_modules = app_loading.haystack_get_app_modules()
        self.assertIsInstance(app_modules, (list, GeneratorType))

        for i in app_modules:
            self.assertIsInstance(i, ModuleType)

    def test_get_models_all(self):
        models = app_loading.haystack_get_models("core")
        self.assertIsInstance(models, (list, GeneratorType))

    def test_get_models_specific(self):
        from test_haystack.core.models import MockModel

        models = app_loading.haystack_get_models("core.MockModel")
        self.assertIsInstance(models, (list, GeneratorType))
        self.assertListEqual(models, [MockModel])

    def test_hierarchal_app_get_models(self):
        models = app_loading.haystack_get_models("hierarchal_app_django")
        self.assertIsInstance(models, (list, GeneratorType))
        self.assertSetEqual(
            set(str(i._meta) for i in models),
            set(
                (
                    "hierarchal_app_django.hierarchalappsecondmodel",
                    "hierarchal_app_django.hierarchalappmodel",
                )
            ),
        )

    def test_hierarchal_app_specific_model(self):
        models = app_loading.haystack_get_models(
            "hierarchal_app_django.HierarchalAppModel"
        )
        self.assertIsInstance(models, (list, GeneratorType))
        self.assertSetEqual(
            set(str(i._meta) for i in models),
            set(("hierarchal_app_django.hierarchalappmodel",)),
        )


class AppWithoutModelsTests(TestCase):
    # Confirm that everything works if an app is enabled

    def test_simple_view(self):
        url = reverse("app-without-models:simple-view")
        resp = self.client.get(url)
        self.assertEqual(resp.content.decode("utf-8"), "OK")



from haystack import indexes

from .models import Checkin


class CheckinSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True)
    username = indexes.CharField(model_attr="username")
    comment = indexes.CharField(model_attr="comment")
    # Again, if you were using GeoDjango, this could be just:
    #   location = indexes.LocationField(model_attr='location')
    location = indexes.LocationField(model_attr="get_location")
    created = indexes.DateTimeField(model_attr="created")

    def get_model(self):
        return Checkin

    def prepare_text(self, obj):
        # Because I don't feel like creating a template just for this.
        return "\n".join([obj.comment, obj.username])



from haystack import indexes

from ..core.models import MockModel, ScoreMockModel


class SimpleMockSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, use_template=True)
    name = indexes.CharField(model_attr="author")
    pub_date = indexes.DateTimeField(model_attr="pub_date")

    def get_model(self):
        return MockModel


class SimpleMockScoreIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, use_template=True)
    score = indexes.CharField(model_attr="score")

    def get_model(self):
        return ScoreMockModel


class SimpleMockUUIDModelIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, model_attr="characteristics")



import datetime
import pickle

from django.test import TestCase
from django.test.utils import override_settings

from haystack import connections, indexes, reset_search_queries
from haystack.backends import SQ, BaseSearchQuery
from haystack.exceptions import FacetingError
from haystack.models import SearchResult
from haystack.query import (
    EmptySearchQuerySet,
    SearchQuerySet,
    ValuesListSearchQuerySet,
    ValuesSearchQuerySet,
)
from haystack.utils.loading import UnifiedIndex
from test_haystack.core.models import (
    AnotherMockModel,
    CharPKMockModel,
    MockModel,
    UUIDMockModel,
)

from .mocks import (
    MOCK_SEARCH_RESULTS,
    CharPKMockSearchBackend,
    MockSearchBackend,
    MockSearchQuery,
    ReadQuerySetMockSearchBackend,
    UUIDMockSearchBackend,
)
from .test_indexes import (
    GhettoAFifthMockModelSearchIndex,
    TextReadQuerySetTestSearchIndex,
)
from .test_views import BasicAnotherMockModelSearchIndex, BasicMockModelSearchIndex


class SQTestCase(TestCase):
    def test_split_expression(self):
        sq = SQ(foo="bar")

        self.assertEqual(sq.split_expression("foo"), ("foo", "content"))
        self.assertEqual(sq.split_expression("foo__exact"), ("foo", "exact"))
        self.assertEqual(sq.split_expression("foo__content"), ("foo", "content"))
        self.assertEqual(sq.split_expression("foo__contains"), ("foo", "contains"))
        self.assertEqual(sq.split_expression("foo__lt"), ("foo", "lt"))
        self.assertEqual(sq.split_expression("foo__lte"), ("foo", "lte"))
        self.assertEqual(sq.split_expression("foo__gt"), ("foo", "gt"))
        self.assertEqual(sq.split_expression("foo__gte"), ("foo", "gte"))
        self.assertEqual(sq.split_expression("foo__in"), ("foo", "in"))
        self.assertEqual(sq.split_expression("foo__startswith"), ("foo", "startswith"))
        self.assertEqual(sq.split_expression("foo__endswith"), ("foo", "endswith"))
        self.assertEqual(sq.split_expression("foo__range"), ("foo", "range"))
        self.assertEqual(sq.split_expression("foo__fuzzy"), ("foo", "fuzzy"))

        # Unrecognized filter. Fall back to exact.
        self.assertEqual(sq.split_expression("foo__moof"), ("foo", "content"))

    def test_repr(self):
        self.assertEqual(repr(SQ(foo="bar")), "<SQ: AND foo__content=bar>")
        self.assertEqual(repr(SQ(foo=1)), "<SQ: AND foo__content=1>")
        self.assertEqual(
            repr(SQ(foo=datetime.datetime(2009, 5, 12, 23, 17))),
            "<SQ: AND foo__content=2009-05-12 23:17:00>",
        )

    def test_simple_nesting(self):
        sq1 = SQ(foo="bar")
        sq2 = SQ(foo="bar")
        bigger_sq = SQ(sq1 & sq2)
        self.assertEqual(
            repr(bigger_sq), "<SQ: AND (foo__content=bar AND foo__content=bar)>"
        )

        another_bigger_sq = SQ(sq1 | sq2)
        self.assertEqual(
            repr(another_bigger_sq), "<SQ: AND (foo__content=bar OR foo__content=bar)>"
        )

        one_more_bigger_sq = SQ(sq1 & ~sq2)
        self.assertEqual(
            repr(one_more_bigger_sq),
            "<SQ: AND (foo__content=bar AND NOT (foo__content=bar))>",
        )

        mega_sq = SQ(bigger_sq & SQ(another_bigger_sq | ~one_more_bigger_sq))
        self.assertEqual(
            repr(mega_sq),
            "<SQ: AND ((foo__content=bar AND foo__content=bar) AND ((foo__content=bar OR foo__content=bar) OR NOT ((foo__content=bar AND NOT (foo__content=bar)))))>",
        )


class BaseSearchQueryTestCase(TestCase):
    fixtures = ["base_data.json", "bulk_data.json"]

    @classmethod
    def setUpClass(cls):
        for connection in connections.all():
            connection.get_unified_index().reset()
        super().setUpClass()

    def setUp(self):
        super().setUp()
        self.bsq = BaseSearchQuery()

    def test_get_count(self):
        self.bsq.add_filter(SQ(foo="bar"))
        self.assertRaises(NotImplementedError, self.bsq.get_count)

    def test_build_query(self):
        self.bsq.add_filter(SQ(foo="bar"))
        self.assertRaises(NotImplementedError, self.bsq.build_query)

    def test_add_filter(self):
        self.assertEqual(len(self.bsq.query_filter), 0)

        self.bsq.add_filter(SQ(foo="bar"))
        self.assertEqual(len(self.bsq.query_filter), 1)

        self.bsq.add_filter(SQ(foo__lt="10"))

        self.bsq.add_filter(~SQ(claris="moof"))

        self.bsq.add_filter(SQ(claris="moof"), use_or=True)

        self.assertEqual(
            repr(self.bsq.query_filter),
            "<SQ: OR ((foo__content=bar AND foo__lt=10 AND NOT (claris__content=moof)) OR claris__content=moof)>",
        )

        self.bsq.add_filter(SQ(claris="moof"))

        self.assertEqual(
            repr(self.bsq.query_filter),
            "<SQ: AND (((foo__content=bar AND foo__lt=10 AND NOT (claris__content=moof)) OR claris__content=moof) AND claris__content=moof)>",
        )

        self.bsq.add_filter(SQ(claris="wtf mate"))

        self.assertEqual(
            repr(self.bsq.query_filter),
            "<SQ: AND (((foo__content=bar AND foo__lt=10 AND NOT (claris__content=moof)) OR claris__content=moof) AND claris__content=moof AND claris__content=wtf mate)>",
        )

    def test_add_order_by(self):
        self.assertEqual(len(self.bsq.order_by), 0)

        self.bsq.add_order_by("foo")
        self.assertEqual(len(self.bsq.order_by), 1)

    def test_clear_order_by(self):
        self.bsq.add_order_by("foo")
        self.assertEqual(len(self.bsq.order_by), 1)

        self.bsq.clear_order_by()
        self.assertEqual(len(self.bsq.order_by), 0)

    def test_add_model(self):
        self.assertEqual(len(self.bsq.models), 0)
        self.assertRaises(AttributeError, self.bsq.add_model, object)
        self.assertEqual(len(self.bsq.models), 0)

        self.bsq.add_model(MockModel)
        self.assertEqual(len(self.bsq.models), 1)

        self.bsq.add_model(AnotherMockModel)
        self.assertEqual(len(self.bsq.models), 2)

    def test_set_limits(self):
        self.assertEqual(self.bsq.start_offset, 0)
        self.assertEqual(self.bsq.end_offset, None)

        self.bsq.set_limits(10, 50)
        self.assertEqual(self.bsq.start_offset, 10)
        self.assertEqual(self.bsq.end_offset, 50)

    def test_clear_limits(self):
        self.bsq.set_limits(10, 50)
        self.assertEqual(self.bsq.start_offset, 10)
        self.assertEqual(self.bsq.end_offset, 50)

        self.bsq.clear_limits()
        self.assertEqual(self.bsq.start_offset, 0)
        self.assertEqual(self.bsq.end_offset, None)

    def test_add_boost(self):
        self.assertEqual(self.bsq.boost, {})

        self.bsq.add_boost("foo", 10)
        self.assertEqual(self.bsq.boost, {"foo": 10})

    def test_add_highlight(self):
        self.assertEqual(self.bsq.highlight, False)

        self.bsq.add_highlight()
        self.assertEqual(self.bsq.highlight, True)

    def test_more_like_this(self):
        mock = MockModel()
        mock.id = 1
        msq = MockSearchQuery()
        msq.backend = MockSearchBackend("mlt")
        ui = connections["default"].get_unified_index()
        bmmsi = BasicMockModelSearchIndex()
        ui.build(indexes=[bmmsi])
        bmmsi.update()
        msq.more_like_this(mock)

        self.assertEqual(msq.get_count(), 23)
        self.assertEqual(int(msq.get_results()[0].pk), MOCK_SEARCH_RESULTS[0].pk)

    def test_add_field_facet(self):
        self.bsq.add_field_facet("foo")
        self.assertEqual(self.bsq.facets, {"foo": {}})

        self.bsq.add_field_facet("bar")
        self.assertEqual(self.bsq.facets, {"foo": {}, "bar": {}})

    def test_add_date_facet(self):
        self.bsq.add_date_facet(
            "foo",
            start_date=datetime.date(2009, 2, 25),
            end_date=datetime.date(2009, 3, 25),
            gap_by="day",
        )
        self.assertEqual(
            self.bsq.date_facets,
            {
                "foo": {
                    "gap_by": "day",
                    "start_date": datetime.date(2009, 2, 25),
                    "end_date": datetime.date(2009, 3, 25),
                    "gap_amount": 1,
                }
            },
        )

        self.bsq.add_date_facet(
            "bar",
            start_date=datetime.date(2008, 1, 1),
            end_date=datetime.date(2009, 12, 1),
            gap_by="month",
        )
        self.assertEqual(
            self.bsq.date_facets,
            {
                "foo": {
                    "gap_by": "day",
                    "start_date": datetime.date(2009, 2, 25),
                    "end_date": datetime.date(2009, 3, 25),
                    "gap_amount": 1,
                },
                "bar": {
                    "gap_by": "month",
                    "start_date": datetime.date(2008, 1, 1),
                    "end_date": datetime.date(2009, 12, 1),
                    "gap_amount": 1,
                },
            },
        )

    def test_add_query_facet(self):
        self.bsq.add_query_facet("foo", "bar")
        self.assertEqual(self.bsq.query_facets, [("foo", "bar")])

        self.bsq.add_query_facet("moof", "baz")
        self.assertEqual(self.bsq.query_facets, [("foo", "bar"), ("moof", "baz")])

        self.bsq.add_query_facet("foo", "baz")
        self.assertEqual(
            self.bsq.query_facets, [("foo", "bar"), ("moof", "baz"), ("foo", "baz")]
        )

    def test_add_stats(self):
        self.bsq.add_stats_query("foo", ["bar"])
        self.assertEqual(self.bsq.stats, {"foo": ["bar"]})

        self.bsq.add_stats_query("moof", ["bar", "baz"])
        self.assertEqual(self.bsq.stats, {"foo": ["bar"], "moof": ["bar", "baz"]})

    def test_add_narrow_query(self):
        self.bsq.add_narrow_query("foo:bar")
        self.assertEqual(self.bsq.narrow_queries, set(["foo:bar"]))

        self.bsq.add_narrow_query("moof:baz")
        self.assertEqual(self.bsq.narrow_queries, set(["foo:bar", "moof:baz"]))

    def test_set_result_class(self):
        # Assert that we're defaulting to ``SearchResult``.
        self.assertTrue(issubclass(self.bsq.result_class, SearchResult))

        # Custom class.
        class IttyBittyResult:
            pass

        self.bsq.set_result_class(IttyBittyResult)
        self.assertTrue(issubclass(self.bsq.result_class, IttyBittyResult))

        # Reset to default.
        self.bsq.set_result_class(None)
        self.assertTrue(issubclass(self.bsq.result_class, SearchResult))

    def test_run(self):
        # Stow.
        self.old_unified_index = connections["default"]._index
        self.ui = UnifiedIndex()
        self.bmmsi = BasicMockModelSearchIndex()
        self.bammsi = BasicAnotherMockModelSearchIndex()
        self.ui.build(indexes=[self.bmmsi, self.bammsi])
        connections["default"]._index = self.ui

        # Update the "index".
        backend = connections["default"].get_backend()
        backend.clear()
        backend.update(self.bmmsi, MockModel.objects.all())

        msq = connections["default"].get_query()
        self.assertEqual(len(msq.get_results()), 23)
        self.assertEqual(int(msq.get_results()[0].pk), MOCK_SEARCH_RESULTS[0].pk)

        # Restore.
        connections["default"]._index = self.old_unified_index

    def test_clone(self):
        self.bsq.add_filter(SQ(foo="bar"))
        self.bsq.add_filter(SQ(foo__lt="10"))
        self.bsq.add_filter(~SQ(claris="moof"))
        self.bsq.add_filter(SQ(claris="moof"), use_or=True)
        self.bsq.add_order_by("foo")
        self.bsq.add_model(MockModel)
        self.bsq.add_boost("foo", 2)
        self.bsq.add_highlight()
        self.bsq.add_field_facet("foo")
        self.bsq.add_date_facet(
            "foo",
            start_date=datetime.date(2009, 1, 1),
            end_date=datetime.date(2009, 1, 31),
            gap_by="day",
        )
        self.bsq.add_query_facet("foo", "bar")
        self.bsq.add_stats_query("foo", "bar")
        self.bsq.add_narrow_query("foo:bar")

        clone = self.bsq._clone()
        self.assertTrue(isinstance(clone, BaseSearchQuery))
        self.assertEqual(len(clone.query_filter), 2)
        self.assertEqual(len(clone.order_by), 1)
        self.assertEqual(len(clone.models), 1)
        self.assertEqual(len(clone.boost), 1)
        self.assertEqual(clone.highlight, True)
        self.assertEqual(len(clone.facets), 1)
        self.assertEqual(len(clone.date_facets), 1)
        self.assertEqual(len(clone.query_facets), 1)
        self.assertEqual(len(clone.narrow_queries), 1)
        self.assertEqual(clone.start_offset, self.bsq.start_offset)
        self.assertEqual(clone.end_offset, self.bsq.end_offset)
        self.assertEqual(clone.backend.__class__, self.bsq.backend.__class__)

    def test_log_query(self):
        reset_search_queries()
        self.assertEqual(len(connections["default"].queries), 0)

        # Stow.
        self.old_unified_index = connections["default"]._index
        self.ui = UnifiedIndex()
        self.bmmsi = BasicMockModelSearchIndex()
        self.ui.build(indexes=[self.bmmsi])
        connections["default"]._index = self.ui

        # Update the "index".
        backend = connections["default"].get_backend()
        backend.clear()
        self.bmmsi.update()

        with self.settings(DEBUG=False):
            msq = connections["default"].get_query()
            self.assertEqual(len(msq.get_results()), 23)
            self.assertEqual(len(connections["default"].queries), 0)

        with self.settings(DEBUG=True):
            # Redefine it to clear out the cached results.
            msq2 = connections["default"].get_query()
            self.assertEqual(len(msq2.get_results()), 23)
            self.assertEqual(len(connections["default"].queries), 1)
            self.assertEqual(connections["default"].queries[0]["query_string"], "")

            msq3 = connections["default"].get_query()
            msq3.add_filter(SQ(foo="bar"))
            len(msq3.get_results())
            self.assertEqual(len(connections["default"].queries), 2)
            self.assertEqual(connections["default"].queries[0]["query_string"], "")
            self.assertEqual(connections["default"].queries[1]["query_string"], "")

        # Restore.
        connections["default"]._index = self.old_unified_index


class CharPKMockModelSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, model_attr="key")

    def get_model(self):
        return CharPKMockModel


class SimpleMockUUIDModelIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, model_attr="characteristics")

    def get_model(self):
        return UUIDMockModel


@override_settings(DEBUG=True)
class SearchQuerySetTestCase(TestCase):
    fixtures = ["base_data.json", "bulk_data.json"]

    def setUp(self):
        super().setUp()

        # Stow.
        self.old_unified_index = connections["default"]._index
        self.ui = UnifiedIndex()
        self.bmmsi = BasicMockModelSearchIndex()
        self.cpkmmsi = CharPKMockModelSearchIndex()
        self.uuidmmsi = SimpleMockUUIDModelIndex()
        self.ui.build(indexes=[self.bmmsi, self.cpkmmsi, self.uuidmmsi])
        connections["default"]._index = self.ui

        # Update the "index".
        backend = connections["default"].get_backend()
        backend.clear()
        backend.update(self.bmmsi, MockModel.objects.all())

        self.msqs = SearchQuerySet()

        # Stow.
        reset_search_queries()

    def tearDown(self):
        # Restore.
        connections["default"]._index = self.old_unified_index
        super().tearDown()

    def test_len(self):
        self.assertEqual(len(self.msqs), 23)

    def test_repr(self):
        reset_search_queries()
        self.assertEqual(len(connections["default"].queries), 0)
        self.assertRegex(
            repr(self.msqs),
            r"^<SearchQuerySet: query=<test_haystack.mocks.MockSearchQuery object"
            r" at 0x[0-9A-Fa-f]+>, using=None>$",
        )

    def test_iter(self):
        reset_search_queries()
        self.assertEqual(len(connections["default"].queries), 0)
        msqs = self.msqs.all()
        results = [int(res.pk) for res in iter(msqs)]
        self.assertEqual(results, [res.pk for res in MOCK_SEARCH_RESULTS[:23]])
        self.assertEqual(len(connections["default"].queries), 3)

    def test_slice(self):
        reset_search_queries()
        self.assertEqual(len(connections["default"].queries), 0)
        results = self.msqs.all()
        self.assertEqual(
            [int(res.pk) for res in results[1:11]],
            [res.pk for res in MOCK_SEARCH_RESULTS[1:11]],
        )
        self.assertEqual(len(connections["default"].queries), 1)

        reset_search_queries()
        self.assertEqual(len(connections["default"].queries), 0)
        results = self.msqs.all()
        self.assertEqual(int(results[22].pk), MOCK_SEARCH_RESULTS[22].pk)
        self.assertEqual(len(connections["default"].queries), 1)

    def test_manual_iter(self):
        results = self.msqs.all()

        reset_search_queries()
        self.assertEqual(len(connections["default"].queries), 0)

        check = [result.pk for result in results._manual_iter()]
        self.assertEqual(
            check,
            [
                "1",
                "2",
                "3",
                "4",
                "5",
                "6",
                "7",
                "8",
                "9",
                "10",
                "11",
                "12",
                "13",
                "14",
                "15",
                "16",
                "17",
                "18",
                "19",
                "20",
                "21",
                "22",
                "23",
            ],
        )

        self.assertEqual(len(connections["default"].queries), 3)

        reset_search_queries()
        self.assertEqual(len(connections["default"].queries), 0)

        # Test to ensure we properly fill the cache, even if we get fewer
        # results back (not a handled model) than the hit count indicates.
        # This will hang indefinitely if broken.

        # CharPK testing
        old_ui = self.ui
        self.ui.build(indexes=[self.cpkmmsi])
        connections["default"]._index = self.ui
        self.cpkmmsi.update()

        results = self.msqs.all()
        loaded = [result.pk for result in results._manual_iter()]
        self.assertEqual(loaded, ["sometext", "1234"])
        self.assertEqual(len(connections["default"].queries), 1)

        # UUID testing
        self.ui.build(indexes=[self.uuidmmsi])
        connections["default"]._index = self.ui
        self.uuidmmsi.update()

        results = self.msqs.all()
        loaded = [result.pk for result in results._manual_iter()]
        self.assertEqual(
            loaded,
            [
                "53554c58-7051-4350-bcc9-dad75eb248a9",
                "77554c58-7051-4350-bcc9-dad75eb24888",
            ],
        )

        connections["default"]._index = old_ui

    def test_cache_is_full(self):
        reset_search_queries()
        self.assertEqual(len(connections["default"].queries), 0)
        self.assertEqual(self.msqs._cache_is_full(), False)
        results = self.msqs.all()
        fire_the_iterator_and_fill_cache = list(results)
        self.assertEqual(23, len(fire_the_iterator_and_fill_cache))
        self.assertEqual(results._cache_is_full(), True)
        self.assertEqual(len(connections["default"].queries), 4)

    def test_all(self):
        sqs = self.msqs.all()
        self.assertTrue(isinstance(sqs, SearchQuerySet))

    def test_filter(self):
        sqs = self.msqs.filter(content="foo")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.query_filter), 1)

    def test_exclude(self):
        sqs = self.msqs.exclude(content="foo")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.query_filter), 1)

    def test_order_by(self):
        sqs = self.msqs.order_by("foo")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertTrue("foo" in sqs.query.order_by)

    def test_models(self):
        # Stow.
        old_unified_index = connections["default"]._index
        ui = UnifiedIndex()
        bmmsi = BasicMockModelSearchIndex()
        bammsi = BasicAnotherMockModelSearchIndex()
        ui.build(indexes=[bmmsi, bammsi])
        connections["default"]._index = ui

        msqs = SearchQuerySet()

        sqs = msqs.all()
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.models), 0)

        sqs = msqs.models(MockModel)
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.models), 1)

        sqs = msqs.models(MockModel, AnotherMockModel)
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.models), 2)

        # This will produce a warning.
        ui.build(indexes=[bmmsi])
        sqs = msqs.models(AnotherMockModel)
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.models), 1)

    def test_result_class(self):
        sqs = self.msqs.all()
        self.assertTrue(issubclass(sqs.query.result_class, SearchResult))

        # Custom class.
        class IttyBittyResult:
            pass

        sqs = self.msqs.result_class(IttyBittyResult)
        self.assertTrue(issubclass(sqs.query.result_class, IttyBittyResult))

        # Reset to default.
        sqs = self.msqs.result_class(None)
        self.assertTrue(issubclass(sqs.query.result_class, SearchResult))

    def test_boost(self):
        sqs = self.msqs.boost("foo", 10)
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.boost.keys()), 1)

    def test_highlight(self):
        sqs = self.msqs.highlight()
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(sqs.query.highlight, True)

    def test_spelling_override(self):
        sqs = self.msqs.filter(content="not the spellchecking query")
        self.assertEqual(sqs.query.spelling_query, None)
        sqs = self.msqs.set_spelling_query("override")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(sqs.query.spelling_query, "override")

    def test_spelling_suggestions(self):
        # Test the case where spelling support is disabled.
        sqs = self.msqs.filter(content="Indx")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(sqs.spelling_suggestion(), None)
        self.assertEqual(sqs.spelling_suggestion("indexy"), None)

    def test_raw_search(self):
        self.assertEqual(len(self.msqs.raw_search("foo")), 23)
        self.assertEqual(
            len(
                self.msqs.raw_search("(content__exact:hello AND content__exact:world)")
            ),
            23,
        )

    def test_load_all(self):
        # Models with character primary keys.
        sqs = SearchQuerySet()
        sqs.query.backend = CharPKMockSearchBackend("charpk")
        results = sqs.load_all().all()
        self.assertEqual(len(results._result_cache), 0)
        results._fill_cache(0, 2)
        self.assertEqual(
            len([result for result in results._result_cache if result is not None]), 2
        )

        # Models with uuid primary keys.
        sqs = SearchQuerySet()
        sqs.query.backend = UUIDMockSearchBackend("uuid")
        results = sqs.load_all().all()
        self.assertEqual(len(results._result_cache), 0)
        results._fill_cache(0, 2)
        self.assertEqual(
            len([result for result in results._result_cache if result is not None]), 2
        )

        # If nothing is handled, you get nothing.
        old_ui = connections["default"]._index
        ui = UnifiedIndex()
        ui.build(indexes=[])
        connections["default"]._index = ui

        sqs = self.msqs.load_all()
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs), 0)

        connections["default"]._index = old_ui

        # For full tests, see the solr_backend.

    def test_load_all_read_queryset(self):
        # Stow.
        old_ui = connections["default"]._index
        ui = UnifiedIndex()
        gafmmsi = GhettoAFifthMockModelSearchIndex()
        ui.build(indexes=[gafmmsi])
        connections["default"]._index = ui
        gafmmsi.update()

        sqs = SearchQuerySet()
        results = sqs.load_all().all()
        results.query.backend = ReadQuerySetMockSearchBackend("default")
        results._fill_cache(0, 2)

        # The deleted result isn't returned
        self.assertEqual(
            len([result for result in results._result_cache if result is not None]), 1
        )

        # Register a SearchIndex with a read_queryset that returns deleted items
        rqstsi = TextReadQuerySetTestSearchIndex()
        ui.build(indexes=[rqstsi])
        rqstsi.update()

        sqs = SearchQuerySet()
        results = sqs.load_all().all()
        results.query.backend = ReadQuerySetMockSearchBackend("default")
        results._fill_cache(0, 2)

        # Both the deleted and not deleted items are returned
        self.assertEqual(
            len([result for result in results._result_cache if result is not None]), 2
        )

        # Restore.
        connections["default"]._index = old_ui

    def test_auto_query(self):
        sqs = self.msqs.auto_query("test search -stuff")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(
            repr(sqs.query.query_filter),
            "<SQ: AND content__content=test search -stuff>",
        )

        sqs = self.msqs.auto_query('test "my thing" search -stuff')
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(
            repr(sqs.query.query_filter),
            '<SQ: AND content__content=test "my thing" search -stuff>',
        )

        sqs = self.msqs.auto_query("test \"my thing\" search 'moar quotes' -stuff")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(
            repr(sqs.query.query_filter),
            "<SQ: AND content__content=test \"my thing\" search 'moar quotes' -stuff>",
        )

        sqs = self.msqs.auto_query('test "my thing" search \'moar quotes\' "foo -stuff')
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(
            repr(sqs.query.query_filter),
            '<SQ: AND content__content=test "my thing" search \'moar quotes\' "foo -stuff>',
        )

        sqs = self.msqs.auto_query("test - stuff")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(
            repr(sqs.query.query_filter), "<SQ: AND content__content=test - stuff>"
        )

        # Ensure bits in exact matches get escaped properly as well.
        sqs = self.msqs.auto_query('"pants:rule"')
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(
            repr(sqs.query.query_filter), '<SQ: AND content__content="pants:rule">'
        )

        # Now with a different fieldname
        sqs = self.msqs.auto_query("test search -stuff", fieldname="title")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(
            repr(sqs.query.query_filter), "<SQ: AND title__content=test search -stuff>"
        )

        sqs = self.msqs.auto_query('test "my thing" search -stuff', fieldname="title")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(
            repr(sqs.query.query_filter),
            '<SQ: AND title__content=test "my thing" search -stuff>',
        )

    def test_count(self):
        self.assertEqual(self.msqs.count(), 23)

    def test_facet_counts(self):
        self.assertEqual(self.msqs.facet_counts(), {})

    def test_best_match(self):
        self.assertTrue(isinstance(self.msqs.best_match(), SearchResult))

    def test_latest(self):
        self.assertTrue(isinstance(self.msqs.latest("pub_date"), SearchResult))

    def test_more_like_this(self):
        mock = MockModel()
        mock.id = 1

        self.assertEqual(len(self.msqs.more_like_this(mock)), 23)

    def test_facets(self):
        sqs = self.msqs.facet("foo")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.facets), 1)

        sqs2 = self.msqs.facet("foo").facet("bar")
        self.assertTrue(isinstance(sqs2, SearchQuerySet))
        self.assertEqual(len(sqs2.query.facets), 2)

    def test_date_facets(self):
        try:
            sqs = self.msqs.date_facet(
                "foo",
                start_date=datetime.date(2008, 2, 25),
                end_date=datetime.date(2009, 2, 25),
                gap_by="smarblaph",
            )
            self.fail()
        except FacetingError as e:
            self.assertEqual(
                str(e),
                "The gap_by ('smarblaph') must be one of the following: year, month, day, hour, minute, second.",
            )

        sqs = self.msqs.date_facet(
            "foo",
            start_date=datetime.date(2008, 2, 25),
            end_date=datetime.date(2009, 2, 25),
            gap_by="month",
        )
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.date_facets), 1)

        sqs2 = self.msqs.date_facet(
            "foo",
            start_date=datetime.date(2008, 2, 25),
            end_date=datetime.date(2009, 2, 25),
            gap_by="month",
        ).date_facet(
            "bar",
            start_date=datetime.date(2007, 2, 25),
            end_date=datetime.date(2009, 2, 25),
            gap_by="year",
        )
        self.assertTrue(isinstance(sqs2, SearchQuerySet))
        self.assertEqual(len(sqs2.query.date_facets), 2)

    def test_query_facets(self):
        sqs = self.msqs.query_facet("foo", "[bar TO *]")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.query_facets), 1)

        sqs2 = self.msqs.query_facet("foo", "[bar TO *]").query_facet(
            "bar", "[100 TO 499]"
        )
        self.assertTrue(isinstance(sqs2, SearchQuerySet))
        self.assertEqual(len(sqs2.query.query_facets), 2)

        # Test multiple query facets on a single field
        sqs3 = (
            self.msqs.query_facet("foo", "[bar TO *]")
            .query_facet("bar", "[100 TO 499]")
            .query_facet("foo", "[1000 TO 1499]")
        )
        self.assertTrue(isinstance(sqs3, SearchQuerySet))
        self.assertEqual(len(sqs3.query.query_facets), 3)

    def test_stats(self):
        sqs = self.msqs.stats_facet("foo", "bar")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.stats), 1)

        sqs2 = self.msqs.stats_facet("foo", "bar").stats_facet("foo", "baz")
        self.assertTrue(isinstance(sqs2, SearchQuerySet))
        self.assertEqual(len(sqs2.query.stats), 1)

        sqs3 = self.msqs.stats_facet("foo", "bar").stats_facet("moof", "baz")
        self.assertTrue(isinstance(sqs3, SearchQuerySet))
        self.assertEqual(len(sqs3.query.stats), 2)

    def test_narrow(self):
        sqs = self.msqs.narrow("foo:moof")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.narrow_queries), 1)

    def test_clone(self):
        results = self.msqs.filter(foo="bar", foo__lt="10")

        clone = results._clone()
        self.assertTrue(isinstance(clone, SearchQuerySet))
        self.assertEqual(str(clone.query), str(results.query))
        self.assertEqual(clone._result_cache, [])
        self.assertEqual(clone._result_count, None)
        self.assertEqual(clone._cache_full, False)
        self.assertEqual(clone._using, results._using)

    def test_using(self):
        sqs = SearchQuerySet(using="default")
        self.assertNotEqual(sqs.query, None)
        self.assertEqual(sqs.query._using, "default")

    def test_chaining(self):
        sqs = self.msqs.filter(content="foo")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.query_filter), 1)

        # A second instance should inherit none of the changes from above.
        sqs = self.msqs.filter(content="bar")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.query_filter), 1)

    def test_none(self):
        sqs = self.msqs.none()
        self.assertTrue(isinstance(sqs, EmptySearchQuerySet))
        self.assertEqual(len(sqs), 0)

    def test___and__(self):
        sqs1 = self.msqs.filter(content="foo")
        sqs2 = self.msqs.filter(content="bar")
        sqs = sqs1 & sqs2

        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.query_filter), 2)

    def test___or__(self):
        sqs1 = self.msqs.filter(content="foo")
        sqs2 = self.msqs.filter(content="bar")
        sqs = sqs1 | sqs2

        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.query_filter), 2)

    def test_and_or(self):
        """
        Combining AND queries with OR should give
            AND(OR(a, b), OR(c, d))
        """
        sqs1 = self.msqs.filter(content="foo").filter(content="oof")
        sqs2 = self.msqs.filter(content="bar").filter(content="rab")
        sqs = sqs1 | sqs2

        self.assertEqual(sqs.query.query_filter.connector, "OR")
        self.assertEqual(
            repr(sqs.query.query_filter.children[0]), repr(sqs1.query.query_filter)
        )
        self.assertEqual(
            repr(sqs.query.query_filter.children[1]), repr(sqs2.query.query_filter)
        )

    def test_or_and(self):
        """
        Combining OR queries with AND should give
            OR(AND(a, b), AND(c, d))
        """
        sqs1 = self.msqs.filter(content="foo").filter_or(content="oof")
        sqs2 = self.msqs.filter(content="bar").filter_or(content="rab")
        sqs = sqs1 & sqs2

        self.assertEqual(sqs.query.query_filter.connector, "AND")
        self.assertEqual(
            repr(sqs.query.query_filter.children[0]), repr(sqs1.query.query_filter)
        )
        self.assertEqual(
            repr(sqs.query.query_filter.children[1]), repr(sqs2.query.query_filter)
        )


class ValuesQuerySetTestCase(SearchQuerySetTestCase):
    def test_values_sqs(self):
        sqs = self.msqs.auto_query("test").values("id")
        self.assertIsInstance(sqs, ValuesSearchQuerySet)

        # We'll do a basic test to confirm that slicing works as expected:
        self.assertIsInstance(sqs[0], dict)
        self.assertIsInstance(sqs[0:5][0], dict)

    def test_valueslist_sqs(self):
        sqs = self.msqs.auto_query("test").values_list("id")

        self.assertIsInstance(sqs, ValuesListSearchQuerySet)
        self.assertIsInstance(sqs[0], (list, tuple))
        self.assertIsInstance(sqs[0:1][0], (list, tuple))

        self.assertRaises(
            TypeError,
            self.msqs.auto_query("test").values_list,
            "id",
            "score",
            flat=True,
        )

        flat_sqs = self.msqs.auto_query("test").values_list("id", flat=True)
        self.assertIsInstance(sqs, ValuesListSearchQuerySet)

        # Note that this will actually be None because a mocked sqs lacks
        # anything else:
        self.assertIsNone(flat_sqs[0])
        self.assertIsNone(flat_sqs[0:1][0])


class EmptySearchQuerySetTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.esqs = EmptySearchQuerySet()

    def test_get_count(self):
        self.assertEqual(self.esqs.count(), 0)
        self.assertEqual(len(self.esqs.all()), 0)

    def test_filter(self):
        sqs = self.esqs.filter(content="foo")
        self.assertTrue(isinstance(sqs, EmptySearchQuerySet))
        self.assertEqual(len(sqs), 0)

    def test_exclude(self):
        sqs = self.esqs.exclude(content="foo")
        self.assertTrue(isinstance(sqs, EmptySearchQuerySet))
        self.assertEqual(len(sqs), 0)

    def test_slice(self):
        sqs = self.esqs.filter(content="foo")
        self.assertTrue(isinstance(sqs, EmptySearchQuerySet))
        self.assertEqual(len(sqs), 0)
        self.assertEqual(sqs[:10], [])

        try:
            sqs[4]
            self.fail()
        except IndexError:
            pass

    def test_dictionary_lookup(self):
        """
        Ensure doing a dictionary lookup raises a TypeError so
        EmptySearchQuerySets can be used in templates.
        """
        self.assertRaises(TypeError, lambda: self.esqs["count"])


@override_settings(DEBUG=True)
class PickleSearchQuerySetTestCase(TestCase):
    fixtures = ["base_data"]

    def setUp(self):
        super().setUp()
        # Stow.
        self.old_unified_index = connections["default"]._index
        self.ui = UnifiedIndex()
        self.bmmsi = BasicMockModelSearchIndex()
        self.cpkmmsi = CharPKMockModelSearchIndex()
        self.ui.build(indexes=[self.bmmsi, self.cpkmmsi])
        connections["default"]._index = self.ui

        # Update the "index".
        backend = connections["default"].get_backend()
        backend.clear()
        backend.update(self.bmmsi, MockModel.objects.all())

        self.msqs = SearchQuerySet()

        # Stow.
        reset_search_queries()

    def tearDown(self):
        # Restore.
        connections["default"]._index = self.old_unified_index
        super().tearDown()

    def test_pickling(self):
        results = self.msqs.all()

        for res in results:
            # Make sure the cache is full.
            pass

        in_a_pickle = pickle.dumps(results)
        like_a_cuke = pickle.loads(in_a_pickle)
        self.assertEqual(len(like_a_cuke), len(results))
        self.assertEqual(like_a_cuke[0].id, results[0].id)



"""Tests for Whoosh spelling suggestions"""

from django.conf import settings
from django.http import HttpRequest

from haystack.forms import SearchForm
from haystack.query import SearchQuerySet
from haystack.views import SearchView

from .test_whoosh_backend import LiveWhooshRoundTripTestCase


class SpellingSuggestionTestCase(LiveWhooshRoundTripTestCase):
    fixtures = ["base_data"]

    def setUp(self):
        self.old_spelling_setting = settings.HAYSTACK_CONNECTIONS["whoosh"].get(
            "INCLUDE_SPELLING", False
        )
        settings.HAYSTACK_CONNECTIONS["whoosh"]["INCLUDE_SPELLING"] = True

        super().setUp()

    def tearDown(self):
        settings.HAYSTACK_CONNECTIONS["whoosh"][
            "INCLUDE_SPELLING"
        ] = self.old_spelling_setting
        super().tearDown()

    def test_form_suggestion(self):
        form = SearchForm({"q": "exampl"}, searchqueryset=SearchQuerySet("whoosh"))
        self.assertEqual(form.get_suggestion(), "example")

    def test_view_suggestion(self):
        view = SearchView(
            template="test_suggestion.html", searchqueryset=SearchQuerySet("whoosh")
        )
        mock = HttpRequest()
        mock.GET["q"] = "exampl"
        resp = view(mock)
        self.assertEqual(resp.content, b"Suggestion: example\n")



import datetime
import unittest

from django.contrib.gis.measure import D
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase

from haystack import connections
from haystack.manager import SearchIndexManager
from haystack.models import SearchResult
from haystack.query import (
    EmptySearchQuerySet,
    SearchQuerySet,
    ValuesListSearchQuerySet,
    ValuesSearchQuerySet,
)
from test_haystack.core.models import MockModel

from .mocks import CharPKMockSearchBackend
from .test_views import BasicAnotherMockModelSearchIndex, BasicMockModelSearchIndex

try:
    from django.contrib.gis.geos import Point

    HAVE_GDAL = True
except ImproperlyConfigured:
    HAVE_GDAL = False


class CustomManager(SearchIndexManager):
    def filter(self, *args, **kwargs):
        return self.get_search_queryset().filter(content="foo1").filter(*args, **kwargs)


class CustomMockModelIndexWithObjectsManager(BasicMockModelSearchIndex):
    objects = CustomManager()


class CustomMockModelIndexWithAnotherManager(BasicMockModelSearchIndex):
    another = CustomManager()


class ManagerTestCase(TestCase):
    fixtures = ["bulk_data.json"]

    def setUp(self):
        super().setUp()

        self.search_index = BasicMockModelSearchIndex
        # Update the "index".
        backend = connections["default"].get_backend()
        backend.clear()
        backend.update(self.search_index(), MockModel.objects.all())
        ui = connections["default"].get_unified_index()
        ui.build([BasicMockModelSearchIndex(), BasicAnotherMockModelSearchIndex()])

        self.search_queryset = BasicMockModelSearchIndex.objects.all()

    def test_queryset(self):
        self.assertTrue(isinstance(self.search_queryset, SearchQuerySet))

    def test_none(self):
        self.assertTrue(
            isinstance(self.search_index.objects.none(), EmptySearchQuerySet)
        )

    def test_filter(self):
        sqs = self.search_index.objects.filter(content="foo")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.query_filter), 1)

    def test_exclude(self):
        sqs = self.search_index.objects.exclude(content="foo")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.query_filter), 1)

    def test_filter_and(self):
        sqs = self.search_index.objects.filter_and(content="foo")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(sqs.query.query_filter.connector, "AND")

    def test_filter_or(self):
        sqs = self.search_index.objects.filter_or(content="foo")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(sqs.query.query_filter.connector, "OR")

    def test_order_by(self):
        sqs = self.search_index.objects.order_by("foo")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertTrue("foo" in sqs.query.order_by)

    @unittest.skipUnless(HAVE_GDAL, "Requires gdal library")
    def test_order_by_distance(self):
        p = Point(1.23, 4.56)
        sqs = self.search_index.objects.distance("location", p).order_by("distance")
        self.assertTrue(isinstance(sqs, SearchQuerySet))

        params = sqs.query.build_params()

        self.assertIn("distance_point", params)
        self.assertDictEqual(
            params["distance_point"], {"field": "location", "point": p}
        )
        self.assertTupleEqual(params["distance_point"]["point"].coords, (1.23, 4.56))

        self.assertListEqual(params["sort_by"], ["distance"])

    def test_highlight(self):
        sqs = self.search_index.objects.highlight()
        self.assertEqual(sqs.query.highlight, True)

    def test_boost(self):
        sqs = self.search_index.objects.boost("foo", 10)
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.boost.keys()), 1)

    def test_facets(self):
        sqs = self.search_index.objects.facet("foo")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.facets), 1)

    @unittest.skipUnless(HAVE_GDAL, "Requires gdal library")
    def test_within(self):
        # This is a meaningless query but we're just confirming that the manager updates the parameters here:
        p1 = Point(-90, -90)
        p2 = Point(90, 90)
        sqs = self.search_index.objects.within("location", p1, p2)
        self.assertTrue(isinstance(sqs, SearchQuerySet))

        params = sqs.query.build_params()

        self.assertIn("within", params)
        self.assertDictEqual(
            params["within"], {"field": "location", "point_1": p1, "point_2": p2}
        )

    @unittest.skipUnless(HAVE_GDAL, "Requires gdal library")
    def test_dwithin(self):
        p = Point(0, 0)
        distance = D(mi=500)
        sqs = self.search_index.objects.dwithin("location", p, distance)
        self.assertTrue(isinstance(sqs, SearchQuerySet))

        params = sqs.query.build_params()

        self.assertIn("dwithin", params)
        self.assertDictEqual(
            params["dwithin"], {"field": "location", "point": p, "distance": distance}
        )

    @unittest.skipUnless(HAVE_GDAL, "Requires gdal library")
    def test_distance(self):
        p = Point(0, 0)
        sqs = self.search_index.objects.distance("location", p)
        self.assertTrue(isinstance(sqs, SearchQuerySet))

        params = sqs.query.build_params()
        self.assertIn("distance_point", params)
        self.assertDictEqual(
            params["distance_point"], {"field": "location", "point": p}
        )

    def test_date_facets(self):
        sqs = self.search_index.objects.date_facet(
            "foo",
            start_date=datetime.date(2008, 2, 25),
            end_date=datetime.date(2009, 2, 25),
            gap_by="month",
        )
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.date_facets), 1)

    def test_query_facets(self):
        sqs = self.search_index.objects.query_facet("foo", "[bar TO *]")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.query_facets), 1)

    def test_narrow(self):
        sqs = self.search_index.objects.narrow("content:foo")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertSetEqual(set(["content:foo"]), sqs.query.narrow_queries)

    def test_raw_search(self):
        self.assertEqual(len(self.search_index.objects.raw_search("foo")), 23)

    def test_load_all(self):
        # Models with character primary keys.
        sqs = self.search_index.objects.all()
        sqs.query.backend = CharPKMockSearchBackend("charpk")
        results = sqs.load_all().all()
        self.assertEqual(len(results._result_cache), 0)

    def test_auto_query(self):
        sqs = self.search_index.objects.auto_query("test search -stuff")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(
            repr(sqs.query.query_filter),
            "<SQ: AND content__content=test search -stuff>",
        )

        # With keyword argument
        sqs = self.search_index.objects.auto_query(
            "test search -stuff", fieldname="title"
        )
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(
            repr(sqs.query.query_filter), "<SQ: AND title__content=test search -stuff>"
        )

    def test_autocomplete(self):
        # Not implemented
        pass

    def test_count(self):
        self.assertEqual(SearchQuerySet().count(), 23)
        self.assertEqual(self.search_index.objects.count(), 23)

    def test_best_match(self):
        self.assertTrue(
            isinstance(self.search_index.objects.best_match(), SearchResult)
        )

    def test_latest(self):
        self.assertTrue(
            isinstance(self.search_index.objects.latest("pub_date"), SearchResult)
        )

    def test_more_like_this(self):
        mock = MockModel()
        mock.id = 1

        self.assertEqual(len(self.search_index.objects.more_like_this(mock)), 23)

    def test_facet_counts(self):
        self.assertEqual(self.search_index.objects.facet_counts(), {})

    def spelling_suggestion(self):
        # Test the case where spelling support is disabled.
        sqs = self.search_index.objects.filter(content="Indx")
        self.assertEqual(sqs.spelling_suggestion(), None)
        self.assertEqual(sqs.spelling_suggestion(preferred_query=None), None)

    def test_values(self):
        sqs = self.search_index.objects.auto_query("test").values("id")
        self.assertIsInstance(sqs, ValuesSearchQuerySet)

    def test_valueslist(self):
        sqs = self.search_index.objects.auto_query("test").values_list("id")
        self.assertIsInstance(sqs, ValuesListSearchQuerySet)


class CustomManagerTestCase(TestCase):
    fixtures = ["bulk_data.json"]

    def setUp(self):
        super().setUp()

        self.search_index_1 = CustomMockModelIndexWithObjectsManager
        self.search_index_2 = CustomMockModelIndexWithAnotherManager

    def test_filter_object_manager(self):
        sqs = self.search_index_1.objects.filter(content="foo")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.query_filter), 2)

    def test_filter_another_manager(self):
        sqs = self.search_index_2.another.filter(content="foo")
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.query_filter), 2)



from haystack import indexes
from haystack.indexes import Indexable, SearchIndex

from .models import Bar, Foo


# To test additional ignores...
class BaseIndex(indexes.SearchIndex):
    text = indexes.CharField(document=True, model_attr="body")

    def get_model(self):
        return Foo


class FooIndex(BaseIndex, indexes.Indexable):
    def index_queryset(self, using=None):
        qs = super().index_queryset(using=using)
        if using == "filtered_whoosh":
            qs = qs.filter(body__contains="1")
        return qs


# Import the old way & make sure things don't explode.


class BarIndex(SearchIndex, Indexable):
    text = indexes.CharField(document=True)

    def get_model(self):
        return Bar

    def prepare_text(self, obj):
        return "%s\n%s" % (obj.author, obj.content)



import os

from django.conf import settings

SECRET_KEY = "CHANGE ME"

# All the normal settings apply. What's included here are the bits you'll have
# to customize.

# Add Haystack to INSTALLED_APPS. You can do this by simply placing in your list.
INSTALLED_APPS = settings.INSTALLED_APPS + ("haystack",)


HAYSTACK_CONNECTIONS = {
    "default": {
        # For Solr:
        "ENGINE": "haystack.backends.solr_backend.SolrEngine",
        "URL": "http://localhost:9001/solr/example",
        "TIMEOUT": 60 * 5,
        "INCLUDE_SPELLING": True,
    },
    "elasticsearch": {
        "ENGINE": "haystack.backends.elasticsearch_backend.ElasticsearchSearchEngine",
        "URL": "http://localhost:9200",
        "INDEX_NAME": "example_project",
    },
    "whoosh": {
        # For Whoosh:
        "ENGINE": "haystack.backends.whoosh_backend.WhooshEngine",
        "PATH": os.path.join(os.path.dirname(__file__), "whoosh_index"),
        "INCLUDE_SPELLING": True,
    },
    "simple": {
        # For Simple:
        "ENGINE": "haystack.backends.simple_backend.SimpleEngine"
    },
    # 'xapian': {
    #     # For Xapian (requires the third-party install):
    #     'ENGINE': 'xapian_backend.XapianEngine',
    #     'PATH': os.path.join(os.path.dirname(__file__), 'xapian_index'),
    # }
}



from django.contrib import admin

from haystack.admin import SearchModelAdmin

from .models import MockModel


@admin.register(MockModel)
class MockModelAdmin(SearchModelAdmin):
    haystack_connection = "solr"
    date_hierarchy = "pub_date"
    list_display = ("author", "pub_date")



from django.test import TestCase

from haystack import connections
from haystack.models import SearchResult
from haystack.query import SQ


class SimpleSearchQueryTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.sq = connections["simple"].get_query()

    def test_build_query_all(self):
        self.assertEqual(self.sq.build_query(), "*")

    def test_build_query_single_word(self):
        self.sq.add_filter(SQ(content="hello"))
        self.assertEqual(self.sq.build_query(), "hello")

    def test_build_query_multiple_word(self):
        self.sq.add_filter(SQ(name="foo"))
        self.sq.add_filter(SQ(name="bar"))
        self.assertEqual(self.sq.build_query(), "foo bar")

    def test_set_result_class(self):
        # Assert that we're defaulting to ``SearchResult``.
        self.assertTrue(issubclass(self.sq.result_class, SearchResult))

        # Custom class.
        class IttyBittyResult:
            pass

        self.sq.set_result_class(IttyBittyResult)
        self.assertTrue(issubclass(self.sq.result_class, IttyBittyResult))

        # Reset to default.
        self.sq.set_result_class(None)
        self.assertTrue(issubclass(self.sq.result_class, SearchResult))



import datetime

from haystack import connections
from haystack.inputs import Exact
from haystack.models import SearchResult
from haystack.query import SQ, SearchQuerySet

from ..core.models import AnotherMockModel, MockModel
from .testcases import WhooshTestCase


class WhooshSearchQueryTestCase(WhooshTestCase):
    def setUp(self):
        super().setUp()

        self.sq = connections["whoosh"].get_query()

    def test_build_query_all(self):
        self.assertEqual(self.sq.build_query(), "*")

    def test_build_query_single_word(self):
        self.sq.add_filter(SQ(content="hello"))
        self.assertEqual(self.sq.build_query(), "(hello)")

    def test_build_query_multiple_words_and(self):
        self.sq.add_filter(SQ(content="hello"))
        self.sq.add_filter(SQ(content="world"))
        self.assertEqual(self.sq.build_query(), "((hello) AND (world))")

    def test_build_query_multiple_words_not(self):
        self.sq.add_filter(~SQ(content="hello"))
        self.sq.add_filter(~SQ(content="world"))
        self.assertEqual(self.sq.build_query(), "(NOT ((hello)) AND NOT ((world)))")

    def test_build_query_multiple_words_or(self):
        self.sq.add_filter(SQ(content="hello") | SQ(content="world"))
        self.assertEqual(self.sq.build_query(), "((hello) OR (world))")

    def test_build_query_multiple_words_mixed(self):
        self.sq.add_filter(SQ(content="why") | SQ(content="hello"))
        self.sq.add_filter(~SQ(content="world"))
        self.assertEqual(
            self.sq.build_query(), "(((why) OR (hello)) AND NOT ((world)))"
        )

    def test_build_query_phrase(self):
        self.sq.add_filter(SQ(content="hello world"))
        self.assertEqual(self.sq.build_query(), "(hello AND world)")

        self.sq.add_filter(SQ(content__exact="hello world"))
        self.assertEqual(
            self.sq.build_query(), '((hello AND world) AND ("hello world"))'
        )

    def test_build_query_boost(self):
        self.sq.add_filter(SQ(content="hello"))
        self.sq.add_boost("world", 5)
        self.assertEqual(self.sq.build_query(), "(hello) world^5")

    def test_correct_exact(self):
        self.sq.add_filter(SQ(content=Exact("hello world")))
        self.assertEqual(self.sq.build_query(), '("hello world")')

    def test_build_query_multiple_filter_types(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(pub_date__lte=datetime.datetime(2009, 2, 10, 1, 59)))
        self.sq.add_filter(SQ(author__gt="daniel"))
        self.sq.add_filter(SQ(created__lt=datetime.datetime(2009, 2, 12, 12, 13)))
        self.sq.add_filter(SQ(title__gte="B"))
        self.sq.add_filter(SQ(id__in=[1, 2, 3]))
        self.sq.add_filter(SQ(rating__range=[3, 5]))
        self.assertEqual(
            self.sq.build_query(),
            "((why) AND pub_date:([to 20090210015900]) AND author:({daniel to}) AND created:({to 20090212121300}) AND title:([B to]) AND id:(1 OR 2 OR 3) AND rating:([3 to 5]))",
        )

    def test_build_query_in_filter_multiple_words(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(title__in=["A Famous Paper", "An Infamous Article"]))
        self.assertEqual(
            self.sq.build_query(),
            '((why) AND title:("A Famous Paper" OR "An Infamous Article"))',
        )

    def test_build_query_in_filter_datetime(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(pub_date__in=[datetime.datetime(2009, 7, 6, 1, 56, 21)]))
        self.assertEqual(self.sq.build_query(), "((why) AND pub_date:(20090706015621))")

    def test_build_query_in_with_set(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(title__in=set(["A Famous Paper", "An Infamous Article"])))
        query = self.sq.build_query()
        self.assertTrue("(why)" in query)

        # Because ordering in Py3 is now random.
        if 'title:("A ' in query:
            self.assertTrue(
                'title:("A Famous Paper" OR "An Infamous Article")' in query
            )
        else:
            self.assertTrue(
                'title:("An Infamous Article" OR "A Famous Paper")' in query
            )

    def test_build_query_wildcard_filter_types(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(title__startswith="haystack"))
        self.assertEqual(self.sq.build_query(), "((why) AND title:(haystack*))")

    def test_build_query_fuzzy_filter_types(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(title__fuzzy="haystack"))
        self.assertEqual(self.sq.build_query(), "((why) AND title:(haystack~2/3))")

    def test_build_query_with_contains(self):
        self.sq.add_filter(SQ(content="circular"))
        self.sq.add_filter(SQ(title__contains="haystack"))
        self.assertEqual(self.sq.build_query(), "((circular) AND title:(*haystack*))")

    def test_build_query_with_endswith(self):
        self.sq.add_filter(SQ(content="circular"))
        self.sq.add_filter(SQ(title__endswith="haystack"))
        self.assertEqual(self.sq.build_query(), "((circular) AND title:(*haystack))")

    def test_clean(self):
        self.assertEqual(self.sq.clean("hello world"), "hello world")
        self.assertEqual(self.sq.clean("hello AND world"), "hello and world")
        self.assertEqual(
            self.sq.clean(
                'hello AND OR NOT TO + - && || ! ( ) { } [ ] ^ " ~ * ? : \ world'
            ),
            "hello and or not to '+' '-' '&&' '||' '!' '(' ')' '{' '}' '[' ']' '^' '\"' '~' '*' '?' ':' '\\' world",
        )
        self.assertEqual(
            self.sq.clean("so please NOTe i am in a bAND and bORed"),
            "so please NOTe i am in a bAND and bORed",
        )

    def test_build_query_with_models(self):
        self.sq.add_filter(SQ(content="hello"))
        self.sq.add_model(MockModel)
        self.assertEqual(self.sq.build_query(), "(hello)")

        self.sq.add_model(AnotherMockModel)
        self.assertEqual(self.sq.build_query(), "(hello)")

    def test_build_query_with_datetime(self):
        self.sq.add_filter(SQ(pub_date=datetime.datetime(2009, 5, 9, 16, 20)))
        self.assertEqual(self.sq.build_query(), "pub_date:(20090509162000)")

    def test_build_query_with_sequence_and_filter_not_in(self):
        self.sq.add_filter(SQ(id=[1, 2, 3]))
        self.assertEqual(self.sq.build_query(), "id:(1,2,3)")

    def test_set_result_class(self):
        # Assert that we're defaulting to ``SearchResult``.
        self.assertTrue(issubclass(self.sq.result_class, SearchResult))

        # Custom class.
        class IttyBittyResult:
            pass

        self.sq.set_result_class(IttyBittyResult)
        self.assertTrue(issubclass(self.sq.result_class, IttyBittyResult))

        # Reset to default.
        self.sq.set_result_class(None)
        self.assertTrue(issubclass(self.sq.result_class, SearchResult))

    def test_in_filter_values_list(self):
        self.sq.add_filter(SQ(content="why"))
        self.sq.add_filter(SQ(title__in=MockModel.objects.values_list("id", flat=True)))
        self.assertEqual(self.sq.build_query(), "((why) AND title:(1 OR 2 OR 3))")

    def test_narrow_sq(self):
        sqs = SearchQuerySet(using="whoosh").narrow(SQ(foo="moof"))
        self.assertTrue(isinstance(sqs, SearchQuerySet))
        self.assertEqual(len(sqs.query.narrow_queries), 1)
        self.assertEqual(sqs.query.narrow_queries.pop(), "foo:(moof)")



from django.db.models import BooleanField, CharField, Model


class HierarchalAppModel(Model):
    enabled = BooleanField(default=True)


class HierarchalAppSecondModel(Model):
    title = CharField(max_length=16)



import logging as std_logging
import pickle

from django.test import TestCase

from haystack import connections
from haystack.models import SearchResult
from haystack.utils import log as logging
from haystack.utils.loading import UnifiedIndex
from test_haystack.core.models import MockModel

from .mocks import MockSearchResult
from .test_indexes import ReadQuerySetTestSearchIndex


class CaptureHandler(std_logging.Handler):
    logs_seen = []

    def emit(self, record):
        CaptureHandler.logs_seen.append(record)


class SearchResultTestCase(TestCase):
    fixtures = ["base_data"]

    def setUp(self):
        super().setUp()
        cap = CaptureHandler()
        logging.getLogger("haystack").addHandler(cap)

        self.no_data = {}
        self.extra_data = {"stored": "I am stored data. How fun."}
        self.no_overwrite_data = {
            "django_id": 2,
            "django_ct": "haystack.anothermockmodel",
            "stored": "I am stored data. How fun.",
        }

        # The str(1) bit might seem unnecessary but it avoids test_unicode needing to handle
        # the differences between repr() output on Python 2 and 3 for a unicode literal:
        self.no_data_sr = MockSearchResult("haystack", "mockmodel", str(1), 2)
        self.extra_data_sr = MockSearchResult(
            "haystack", "mockmodel", str(1), 3, **self.extra_data
        )
        self.no_overwrite_data_sr = MockSearchResult(
            "haystack", "mockmodel", str(1), 4, **self.no_overwrite_data
        )

    def test_init(self):
        self.assertEqual(self.no_data_sr.app_label, "haystack")
        self.assertEqual(self.no_data_sr.model_name, "mockmodel")
        self.assertEqual(self.no_data_sr.model, MockModel)
        self.assertEqual(self.no_data_sr.verbose_name, "Mock model")
        self.assertEqual(self.no_data_sr.verbose_name_plural, "Mock models")
        self.assertEqual(self.no_data_sr.pk, "1")
        self.assertEqual(self.no_data_sr.score, 2)
        self.assertEqual(self.no_data_sr.stored, None)

        self.assertEqual(self.extra_data_sr.app_label, "haystack")
        self.assertEqual(self.extra_data_sr.model_name, "mockmodel")
        self.assertEqual(self.extra_data_sr.model, MockModel)
        self.assertEqual(self.extra_data_sr.verbose_name, "Mock model")
        self.assertEqual(self.extra_data_sr.verbose_name_plural, "Mock models")
        self.assertEqual(self.extra_data_sr.pk, "1")
        self.assertEqual(self.extra_data_sr.score, 3)
        self.assertEqual(self.extra_data_sr.stored, "I am stored data. How fun.")

        self.assertEqual(self.no_overwrite_data_sr.app_label, "haystack")
        self.assertEqual(self.no_overwrite_data_sr.model_name, "mockmodel")
        self.assertEqual(self.no_overwrite_data_sr.model, MockModel)
        self.assertEqual(self.no_overwrite_data_sr.verbose_name, "Mock model")
        self.assertEqual(self.no_overwrite_data_sr.verbose_name_plural, "Mock models")
        self.assertEqual(self.no_overwrite_data_sr.pk, "1")
        self.assertEqual(self.no_overwrite_data_sr.score, 4)
        self.assertEqual(self.no_overwrite_data_sr.stored, "I am stored data. How fun.")

    def test_get_additional_fields(self):
        self.assertEqual(self.no_data_sr.get_additional_fields(), {})
        self.assertEqual(
            self.extra_data_sr.get_additional_fields(),
            {"stored": "I am stored data. How fun."},
        )
        self.assertEqual(
            self.no_overwrite_data_sr.get_additional_fields(),
            {
                "django_ct": "haystack.anothermockmodel",
                "django_id": 2,
                "stored": "I am stored data. How fun.",
            },
        )

    def test_unicode(self):
        self.assertEqual(
            self.no_data_sr.__str__(), "<SearchResult: haystack.mockmodel (pk='1')>"
        )
        self.assertEqual(
            self.extra_data_sr.__str__(),
            "<SearchResult: haystack.mockmodel (pk='1')>",
        )
        self.assertEqual(
            self.no_overwrite_data_sr.__str__(),
            "<SearchResult: haystack.mockmodel (pk='1')>",
        )

    def test_content_type(self):
        self.assertEqual(self.no_data_sr.content_type(), "core.mockmodel")
        self.assertEqual(self.extra_data_sr.content_type(), "core.mockmodel")
        self.assertEqual(self.no_overwrite_data_sr.content_type(), "core.mockmodel")

    def test_stored_fields(self):
        # Stow.
        old_unified_index = connections["default"]._index
        ui = UnifiedIndex()
        ui.build(indexes=[])
        connections["default"]._index = ui

        # Without registering, we should receive an empty dict.
        self.assertEqual(self.no_data_sr.get_stored_fields(), {})
        self.assertEqual(self.extra_data_sr.get_stored_fields(), {})
        self.assertEqual(self.no_overwrite_data_sr.get_stored_fields(), {})

        from haystack import indexes

        class TestSearchIndex(indexes.SearchIndex, indexes.Indexable):
            stored = indexes.CharField(model_attr="author", document=True)

            def get_model(self):
                return MockModel

        # Include the index & try again.
        ui.document_field = "stored"
        ui.build(indexes=[TestSearchIndex()])

        self.assertEqual(self.no_data_sr.get_stored_fields(), {"stored": None})
        self.assertEqual(
            self.extra_data_sr.get_stored_fields(),
            {"stored": "I am stored data. How fun."},
        )
        self.assertEqual(
            self.no_overwrite_data_sr.get_stored_fields(),
            {"stored": "I am stored data. How fun."},
        )

        # Restore.
        connections["default"]._index = old_unified_index

    def test_missing_object(self):
        awol1 = SearchResult("core", "mockmodel", "1000000", 2)
        self.assertEqual(awol1.app_label, "core")
        self.assertEqual(awol1.model_name, "mockmodel")
        self.assertEqual(awol1.pk, "1000000")
        self.assertEqual(awol1.score, 2)

        awol2 = SearchResult("core", "yetanothermockmodel", "1000000", 2)
        self.assertEqual(awol2.app_label, "core")
        self.assertEqual(awol2.model_name, "yetanothermockmodel")
        self.assertEqual(awol2.pk, "1000000")
        self.assertEqual(awol2.score, 2)

        # Failed lookups should fail gracefully.
        CaptureHandler.logs_seen = []
        self.assertEqual(awol1.model, MockModel)
        self.assertEqual(awol1.object, None)
        self.assertEqual(awol1.verbose_name, "Mock model")
        self.assertEqual(awol1.verbose_name_plural, "Mock models")
        self.assertEqual(awol1.stored, None)
        self.assertEqual(len(CaptureHandler.logs_seen), 4)

        CaptureHandler.logs_seen = []
        self.assertEqual(awol2.model, None)
        self.assertEqual(awol2.object, None)
        self.assertEqual(awol2.verbose_name, "")
        self.assertEqual(awol2.verbose_name_plural, "")
        self.assertEqual(awol2.stored, None)
        self.assertEqual(len(CaptureHandler.logs_seen), 12)

    def test_read_queryset(self):
        # The model is flagged deleted so not returned by the default manager.
        deleted1 = SearchResult("core", "afifthmockmodel", 2, 2)
        self.assertEqual(deleted1.object, None)

        # Stow.
        old_unified_index = connections["default"]._index
        ui = UnifiedIndex()
        ui.document_field = "author"
        ui.build(indexes=[ReadQuerySetTestSearchIndex()])
        connections["default"]._index = ui

        # The soft delete manager returns the object.
        deleted2 = SearchResult("core", "afifthmockmodel", 2, 2)
        self.assertNotEqual(deleted2.object, None)
        self.assertEqual(deleted2.object.author, "sam2")

        # Restore.
        connections["default"]._index = old_unified_index

    def test_pickling(self):
        pickle_me_1 = SearchResult("core", "mockmodel", "1000000", 2)
        picklicious = pickle.dumps(pickle_me_1)

        pickle_me_2 = pickle.loads(picklicious)
        self.assertEqual(pickle_me_1.app_label, pickle_me_2.app_label)
        self.assertEqual(pickle_me_1.model_name, pickle_me_2.model_name)
        self.assertEqual(pickle_me_1.pk, pickle_me_2.pk)
        self.assertEqual(pickle_me_1.score, pickle_me_2.score)



import os

from ..utils import check_solr


def load_tests(loader, standard_tests, pattern):
    check_solr()
    package_tests = loader.discover(
        start_dir=os.path.dirname(__file__), pattern=pattern
    )
    standard_tests.addTests(package_tests)
    return standard_tests
